import random
from typing import Optional, Union


class GeneticAlgorithm:
    r"""An implementation of a genetic algorithm. Takes a :attr:`target` list
    and a list of :attr:`genes` to prod`uce generations of individuals at each
    :meth:`__call__`. Fittest individual of the current generation can be
    accessed directly using the property :attr:`fittest_individual`.

    Basic usage:
        The genetic algorithm should be initialised with two :obj:`list`'s of
        objects. The elements of these :obj:`list` can be of any type. The
        first one is the :attr:`target` list used to evaluate the individuals
        in the population. The second is the :attr:`genes` list which should
        contain all possible elements that can make up an individual. All
        elements of :attr:`target` must also be present in :attr:`genes`. For
        this example, :attr:`population_size` will be set to a very small value
        of ``4``, though in practice it's better to use substantially larger
        numbers (default value is ``100``). :attr:`select_n_parents` also needs
        to be decreased, as it must always be smaller than
        :attr:`population_size`.

        >>> ga = auxjad.GeneticAlgorithm(
        ...     target=['A', 'B', 'C', 'D', 'E'],
        ...     genes=['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J'],
        ...     population_size=4,
        ...     select_n_parents=2,
        ... )

        Calling the GA generator will generate a :attr:`population` of
        individuals, each of wich will also be awarded a score according to an
        evaluation function. :attr:`population` and :attr:`scores` are always
        ordered from the fittest to the least fit.

        >>> ga()
        >>> ga.population
        [['A', 'J', 'J', 'B', 'H'],
         ['D', 'A', 'E', 'A', 'F'],
         ['F', 'F', 'A', 'F', 'F'],
         ['F', 'F', 'E', 'J', 'C'],
         ]
        >>> ga.scores
        [0.209603072,
         0.0912,
         0.05638400000000001,
         0.016396800000000003,
         ]

        Each subsequent call will generate a new generation of individuals
        which will become increasingly fit.

    :attr:`fittest_individual` and :attr:`fittest_individual_score`:
        The fittest individual of the whole population can be directly
        accessed using the property :attr:`fittest_individual`:

        >>> ga = auxjad.GeneticAlgorithm(
        ...     target=['A', 'B', 'C', 'D', 'E'],
        ...     genes=['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J'],
        ...     population_size=4,
        ...     select_n_parents=2,
        ... )
        >>> ga()
        >>> ga.fittest_individual
        ['A', 'J', 'J', 'B', 'H']

        Its score is also directly accessible using
        :attr:`fittest_individual_score`:

        >>> ga.fittest_individual_score
        0.209603072

    Evolution:
        As expected, each generation will become increasingly fitter in
        relation to the :attr:`target`. The example below shows the fittest
        individual of each of ``10`` generations, with a
        :attr:`population_size` to  ``50``:

        >>> ga = auxjad.GeneticAlgorithm(
        ...     target=['A', 'B', 'C', 'D', 'E'],
        ...     genes=['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J'],
        ...     population_size=50,
        ... )
        >>> for _ in range(10):
        ...     ga()
        ...     print(ga.fittest_individual, ga.fittest_individual_score)
        ['A', 'J', 'B', 'E', 'E'] 0.480000512
        ['A', 'H', 'C', 'D', 'J'] 0.6000768
        ['A', 'C', 'D', 'D', 'E'] 0.6799999999999999
        ['A', 'C', 'D', 'D', 'E'] 0.6799999999999999
        ['A', 'C', 'D', 'D', 'E'] 0.6799999999999999
        ['A', 'C', 'C', 'D', 'E'] 0.8400000000000001
        ['A', 'C', 'C', 'D', 'E'] 0.8400000000000001
        ['A', 'C', 'C', 'D', 'E'] 0.8400000000000001
        ['A', 'C', 'C', 'D', 'E'] 0.8400000000000001
        ['A', 'C', 'C', 'D', 'E'] 0.8400000000000001

    Evaluation function and :attr:`evaluation_index`:
        The evaluation function gives out a score between ``0.0`` to ``1.0`` to
        each individual in the population. For each given individual, this
        function compares each element of its elements with the element in the
        :attr:`target` at the same index. As to accomodate arbitrary objects
        being used as genes (as opposed to numbers only), this comparison uses
        the distance between the current individual and target genes using the
        :attr:`genes` property. Consider the following example, where the
        available genes are ``['A', 'B', 'C', 'D', 'E', 'F']`` and the target
        is ``['B', 'A', 'A', 'C']``. Suppose an individual has the genes
        ``['D', 'D', 'A', 'B']``.

        To evaluate this individual, first the algorithm finds the indices of
        both the target's genes (in this case, ``[1, 0, 0, 2]``) and also of
        the individual to be evaluated (in this case, ``[3, 3, 0, 1]``). It
        then scores each element of this individual against the target using:

        .. code-block::

            difference = abs(target_gene_index - individual_gene_index)
            element_score = evaluation_index ** difference

        Thus, when the difference is ``0``, the score of this element is
        ``1.0``. The higher the difference, the smaller the value; this decay
        can be controlled using the property :attr:`evaluation_index`, whose
        default value is ``0.2``. Thus when the difference is ``1`` or ``-1``,
        the score is ``0.2 ** 1 = 0.2``, when the difference is ``2`` or
        ``-2``, the score is ``0.2 ** 2 = 0.04``, and so on. The total score of
        an individual will be given by the normalised sum of the evaluation of
        each of its genes.

    :attr:`select_n_parents` and :attr:`keep_n_parents`:
        When creating a new generation of individuals, the algorithm will
        select the fittest individuals to be parents for the next generation.
        The number of parents is given by :attr:`select_n_parents`, whose
        default value is ``10``. These parents generate offspring by being
        recombined through a crossover process, in which the first half of a
        random parent and a second half of another random parent are combined
        into a new individual. Parents are not kept in the next generation by
        default and are solely used in this crossover process. To 'clone'
        parents into the next generation, set the property
        :attr:`keep_n_parents` to a value larger than ``0`` (its default
        value).

    :attr:`mutation_chance` and :attr:`mutation_index`:
        After the crossover process, an individual can go through random
        mutations. The chance of this occuring on a given individual is given
        by :attr:`mutation_chance`, whose value should range between ``0.0``
        and ``1.0`` (default is ``0.2``). If an individual is selected to
        undergo a random mutation, then each of its genes will have a chance of
        being mutated, which in turn is given by the property
        :attr:`mutation_index` (default value is ``0.1``).

    :attr:`initial_individual`:
        Instead of starting with a random population of individuals, it's
        possible to set a custom initial individual as a starting point. This
        can be used to 'morph' fittest individual from this initial one to a
        fit one according to the target.

        >>> ga = auxjad.GeneticAlgorithm(
        ...     target=['A', 'B', 'C', 'D', 'E'],
        ...     genes=['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J'],
        ...     initial_individual=['F', 'G', 'H', 'I', 'J'],
        ... )
        >>> for _ in range(10):
        ...     ga()
        ...     ga.fittest_individual
        ['F', 'G', 'H', 'I', 'J']
        ['F', 'B', 'H', 'I', 'J']
        ['F', 'B', 'H', 'I', 'E']
        ['F', 'B', 'H', 'I', 'E']
        ['A', 'B', 'H', 'I', 'E']
        ['A', 'B', 'E', 'I', 'E']
        ['A', 'B', 'E', 'D', 'E']
        ['A', 'B', 'E', 'D', 'E']
        ['A', 'B', 'E', 'D', 'E']
        ['A', 'B', 'D', 'D', 'E']

    :meth:`reset`:
        Use the :meth:`reset` method to reset the genetic algorithm at any
        point:

        >>> ga = auxjad.GeneticAlgorithm(
        ...     target=['A', 'B', 'C', 'D', 'E'],
        ...     genes=['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J'],
        ... )
        >>> ga.generation_number
        None
        >>> ga.fittest_individual
        None
        >>> for _ in range(10):
        ...     ga()
        >>> ga.generation_number
        9
        >>> ga.fittest_individual
        ['C', 'B', 'C', 'D', 'E']
        >>> ga.reset()
        >>> ga.generation_number
        None
        >>> ga.fittest_individual
        None
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_target",
        "_genes",
        "_initial_individual",
        "_population_size",
        "_select_n_parents",
        "_keep_n_parents",
        "_mutation_chance",
        "_mutation_index",
        "_evaluation_index",
        "_target_indices",
        "_scores",
        "_generation_number",
        "_population",
    )

    ### INITIALISER ###

    def __init__(
        self,
        *,
        target: list,
        genes: list,
        initial_individual: Optional[list] = None,
        population_size: int = 100,
        select_n_parents: int = 10,
        keep_n_parents: int = 0,
        mutation_chance: float = 0.2,
        mutation_index: float = 0.1,
        evaluation_index: float = 0.2,
    ) -> None:
        if not isinstance(genes, list):
            raise TypeError("'genes' must be 'list'")
        if not isinstance(target, list):
            raise TypeError("'target' must be 'list'")
        if any(value not in genes for value in target):
            raise ValueError("'target' must only contain elements present in 'genes'")
        self._genes = genes
        self._target = target
        self._target_indices = [self._genes.index(gene) for gene in self._target]
        self.initial_individual = initial_individual
        self.population_size = population_size
        self.select_n_parents = select_n_parents
        self.keep_n_parents = keep_n_parents
        self.mutation_chance = mutation_chance
        self.mutation_index = mutation_index
        self.evaluation_index = evaluation_index
        self._generation_number = None
        self._population = None
        self._scores = None

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r"""Returns interpreter representation of :attr:`target`."""
        return repr(self._target)

    def __len__(self) -> int:
        r"""Returns the number of genes in each individual."""
        return len(self._target)

    def __call__(self) -> None:
        r"""Calls the genetic algorithm process for one iteration. Creates a
        new generation of length :attr:`population_size` via reproduction and
        mutation processes and scores each individual using the evaluation
        function. Sorts the population according to their scores.
        """
        self._generate_population()
        self._score_population()
        self._sort_population_by_evaluation()

    def __next__(self) -> None:
        r"""Calls the genetic algorithm process for one iteration. Creates a
        new generation of length :attr:`population_size` via reproduction and
        mutation processes and scores each individual using the evaluation
        function. Sorts the population according to their scores.
        """
        try:
            return self.__call__()
        except RuntimeError:
            raise StopIteration

    def __iter__(self) -> None:
        r"""Returns an iterator, allowing instances to be used as iterators."""
        return self

    ### PUBLIC METHODS ###

    def reset(self) -> None:
        r"""Resets that genetic algorithm."""
        self._generation_number = None
        self._population = None
        self._scores = None

    ### PRIVATE METHODS ###

    def _generate_population(self) -> None:
        r"""Calls the genetic algorithm process for one iteration. Creates a
        new generation of length :attr:`population_size` via reproduction and
        mutation processes and scores each individual using the evaluation
        function.
        """
        if self._generation_number is None:
            self._generation_number = 0
            self._generate_initial_individual()
        else:
            self._generation_number += 1
            self._crossover_population()
            self._mutate_population()

    def _generate_initial_individual(self) -> None:
        r"""Generates a random initial population of size
        :attr:`population_size` and whose genes are randomly chosen from
        :attr:`genes`.
        """
        self._population = []
        for _ in range(self._population_size):
            if self._initial_individual is None:
                individual = [random.choice(self._genes) for _ in range(self.__len__())]
            else:
                individual = self._initial_individual[:]
            self._population.append(individual)

    def _evaluate(
        self,
        individual: list,
    ) -> float:
        r"""Evaluates all genes of a given individual, returning a
        :obj:`float`. The higher the value, the fitter the individual is, with
        ``1.0`` being a perfect fit. Use :attr:`evaluation_index` to tweak the
        behaviour of this function.
        """
        individual_score = 0
        individual_indices = [self.genes.index(gene) for gene in individual]
        for gene, target_gene in zip(individual_indices, self._target_indices):
            difference = abs(gene - target_gene)
            individual_score += self._evaluation_index**difference
        individual_score /= self.__len__()
        return individual_score

    def _sort_population_by_evaluation(self) -> None:
        r"""Sorts the population (and their scores) according to the evaluation
        of its individuals.
        """
        zipped_lists = list(zip(self._scores, self._population))
        zipped_lists.sort(
            key=lambda tuple_pair: tuple_pair[0],
            reverse=True,
        )
        self._population = [gene for _, gene in zipped_lists]
        self._scores = [score for score, _ in zipped_lists]

    def _score_population(self) -> None:
        r"""Generates the list of score for each individual of the current
        generation.
        """
        self._scores = []
        for individual in self._population:
            score = self._evaluate(individual)
            self._scores.append(score)

    def _crossover_population(self) -> None:
        r"""Crossover process used to generate offsprings."""
        selected_parents = self._population[: self._select_n_parents]
        if self._keep_n_parents > 0:
            new_generation = self._population[: self._keep_n_parents]
        else:
            new_generation = []
        for _ in range(self._population_size - self._keep_n_parents):
            parents = random.sample(
                range(len(selected_parents)),
                k=2,
            )
            half_index = int(self.__len__() / 2)
            parent_A = selected_parents[parents[0]]
            parent_B = selected_parents[parents[1]]
            individual = parent_A[:half_index] + parent_B[half_index:]
            new_generation.append(individual)
        self._population = new_generation[:]

    def _mutate_population(self) -> None:
        r"""Mutates some individuals of the current generation according to
        :attr:`mutation_chance` and :attr:`mutation_index`.
        """
        for index, individual in enumerate(self._population):
            if random.random() < self._mutation_chance:
                mutated_individual = individual[:]
                for i in range(self.__len__()):
                    if random.random() < self._mutation_index:
                        mutated_individual[i] = random.choice(self._genes)
                self._population[index] = mutated_individual

    ### PUBLIC PROPERTIES ###

    @property
    def target(self) -> list:
        r"""Target individual used for evaluation."""
        return self._target

    @target.setter
    def target(
        self,
        target: list,
    ) -> None:
        if not isinstance(target, list):
            raise TypeError("'target' must be 'list'")
        if any(value not in self._genes for value in target):
            raise ValueError("'target' must only contain elements present in 'genes'")
        self._target = target
        self._target_indices = [self._genes.index(gene) for gene in self._target]

    @property
    def genes(self) -> list:
        r"""List of possible genes that make up all individuals."""
        return self._genes

    @genes.setter
    def genes(
        self,
        genes: list,
    ) -> None:
        if not isinstance(genes, list):
            raise TypeError("'genes' must be 'list'")
        self._genes = genes
        self._target_indices = [self._genes.index(gene) for gene in self._target]

    @property
    def initial_individual(self) -> Union[list, None]:
        r"""Optional initial individual (instead of random initial
        population).
        """
        return self._initial_individual

    @initial_individual.setter
    def initial_individual(
        self,
        initial_individual: Optional[list],
    ) -> None:
        if initial_individual is not None:
            if not isinstance(initial_individual, list):
                raise TypeError("'initial_individual' must be 'list' or 'None'")
            if len(initial_individual) != self.__len__():
                raise ValueError("'initial_individual' must have the same length as 'target'")
            if any(value not in self._genes for value in initial_individual):
                raise ValueError(
                    "'initial_individual' must only contain elements present in 'genes'"
                )
        self._initial_individual = initial_individual

    @property
    def population_size(self) -> int:
        r"""Number of individuals in any given generation."""
        return self._population_size

    @population_size.setter
    def population_size(
        self,
        population_size: int,
    ) -> None:
        if not isinstance(population_size, int):
            raise TypeError("'population_size' must be 'int'")
        self._population_size = population_size

    @property
    def select_n_parents(self) -> int:
        r"""Number of the best-fit individuals that are selected to be the
        parents for the next generation.
        """
        return self._select_n_parents

    @select_n_parents.setter
    def select_n_parents(
        self,
        select_n_parents: int,
    ) -> None:
        if not isinstance(select_n_parents, int):
            raise TypeError("'select_n_parents' must be 'int'")
        if select_n_parents > self._population_size:
            raise ValueError(
                "'select_n_parents' must be equal to or smaller than 'population_size'"
            )
        self._select_n_parents = select_n_parents

    @property
    def keep_n_parents(self) -> int:
        r"""Number of the best-fit individuals that survive into the next
        generation. Default is ``0``.
        """
        return self._keep_n_parents

    @keep_n_parents.setter
    def keep_n_parents(
        self,
        keep_n_parents: int,
    ) -> None:
        if not isinstance(keep_n_parents, int):
            raise TypeError("'keep_n_parents' must be 'int'")
        if keep_n_parents > self._population_size:
            raise ValueError("'keep_n_parents' must be equal to or smaller than 'population_size'")
        self._keep_n_parents = keep_n_parents

    @property
    def mutation_chance(self) -> float:
        r"""The chance of any given individual experiencing mutation."""
        return self._mutation_chance

    @mutation_chance.setter
    def mutation_chance(
        self,
        mutation_chance: float,
    ) -> None:
        if not isinstance(mutation_chance, float):
            raise TypeError("'mutation_chance' must be 'float'")
        if mutation_chance < 0.0:
            raise ValueError("'mutation_chance' must be a positive 'float'")
        elif mutation_chance > 1.0:
            raise ValueError("'mutation_chance' cannot be larger than 1.0")
        self._mutation_chance = mutation_chance

    @property
    def mutation_index(self) -> float:
        r"""Given an individual selected to undergo mutation, this index gives
        the percentage of genes of that individual which will be mutated.
        """
        return self._mutation_index

    @mutation_index.setter
    def mutation_index(
        self,
        mutation_index: float,
    ) -> None:
        if not isinstance(mutation_index, float):
            raise TypeError("'mutation_index' must be 'float'")
        if mutation_index < 0.0:
            raise ValueError("'mutation_index' must be a positive 'float'")
        elif mutation_index > 1.0:
            raise ValueError("'mutation_index' cannot be larger than 1.0")
        self._mutation_index = mutation_index

    @property
    def evaluation_index(self) -> float:
        r"""The index used in the evaluation function. This index will be
        raised by the difference between indices of the target value and the
        current value. Consider the following example, where the available
        genes are ``['A', 'B', 'C', 'D', 'E', 'F']`` and the target is
        ``['B', 'A', 'A', 'C']``. Suppose an individual has the genes
        ``['D', 'D', 'A', 'B']``.

        To evaluate this individual, first the algorithm finds the indices of
        both the target's genes (in this case, ``[1, 0, 0, 2]``) and also of
        the individual to be evaluated (in this case, ``[3, 3, 0, 1]``). It
        then scores each element of this individual against the target using:

        .. code-block::

            difference = abs(target_gene_index - individual_gene_index)
            element_score = evaluation_index ** difference

        Thus, when the difference is ``0``, the score of this element is
        ``1.0``. The higher the difference, the smaller the value; this decay
        can be controlled using this very property, whose default value is
        ``0.2``. Thus when the difference is ``1`` or ``-1``, the score is
        ``0.2 ** 1 = 0.2``, when the difference is ``2`` or ``-2``, the score
        is ``0.2 ** 2 = 0.04``, and so on. The total score of an individual
        will be given by the normalised sum of the evaluation of each of its
        genes.
        """
        return self._evaluation_index

    @evaluation_index.setter
    def evaluation_index(
        self,
        evaluation_index: float,
    ) -> None:
        if not isinstance(evaluation_index, float):
            raise TypeError("'evaluation_index' must be 'float'")
        if evaluation_index <= 0.0:
            raise ValueError("'evaluation_index' must be a greater than 0.0")
        elif evaluation_index >= 1.0:
            raise ValueError("'evaluation_index' must be less than 1.0")
        self._evaluation_index = evaluation_index

    @property
    def generation_number(self) -> Union[int, None]:
        r"""Read-only property, returns the number of the current generation
        (initial generation is ``0``).
        """
        return self._generation_number

    @property
    def population(self) -> Union[list, None]:
        r"""Read-only property, returns a list with all the population of the
        current generation.
        """
        return self._population

    @property
    def scores(self) -> list:
        r"""Read-only property, returns the list of individual scores of the
        current population. Scores are normalised.
        """
        return self._scores

    @property
    def fittest_individual(self) -> Union[list, None]:
        r"""Read-only property, returns the fittest individual of the current
        population.
        """
        try:
            return self._population[0]
        except TypeError:
            return None

    @property
    def fittest_individual_score(self) -> Union[list, float]:
        r"""Read-only property, returns the score of the fittest individual of
        the current population.
        """
        try:
            return self._scores[0]
        except TypeError:
            return None



import abjad

from .underfull_duration import underfull_duration


def selection_is_full(selection: abjad.Selection) -> bool:
    r"""Returns a :obj:`bool` representing whether the last measure of an input
    |abjad.Selection| is fully filled in or not.

    Basic usage:
        Returns ``True`` if the last measure of a selection is full, otherwise
        returns ``False``. If no time signature is encountered at the
        beginning, it uses LilyPond's convention and considers the container
        as in 4/4.

        >>> container1 = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Container(r"c'4 d'4 e'4")
        >>> container3 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4")
        >>> container4 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4 d'4 e'4 f'4")
        >>> auxjad.get.selection_is_full(container1[:])
        True
        >>> auxjad.get.selection_is_full(container2[:])
        False
        >>> auxjad.get.selection_is_full(container3[:])
        False
        >>> auxjad.get.selection_is_full(container4[:])
        True

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.get|. It can thus be used from either |auxjad.get|_ or
        |abjad.get| namespaces. Therefore, the two lines below are equivalent:

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> auxjad.get.selection_is_full(container[:])
        True
        >>> abjad.get.selection_is_full(container[:])
        True

    Time signature changes:
        Handles any time signatures as well as changes of time signature.

        >>> container1 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Container(r"\time 3/4 a2. \time 2/4 r2")
        >>> container3 = abjad.Container(r"\time 5/4 g1 ~ g4 \time 4/4 af'2")
        >>> container4 = abjad.Container(r"\time 6/8 c'2 ~ c'8")
        >>> auxjad.get.selection_is_full(container1[:])
        True
        >>> auxjad.get.selection_is_full(container2[:])
        True
        >>> auxjad.get.selection_is_full(container3[:])
        False
        >>> auxjad.get.selection_is_full(container4[:])
        False

    Partial time signatures:
        Correctly handles partial time signatures.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> time_signature = abjad.TimeSignature((3, 4), partial=(1, 4))
        >>> abjad.attach(time_signature, container[0])
        >>> auxjad.get.selection_is_full(container[:])
        True

    Multi-measure rests:
        It also handles multi-measure rests.

        >>> container1 = abjad.Container(r"R1")
        >>> container2 = abjad.Container(r"\time 3/4 R1 * 3/4 \time 2/4 r2")
        >>> container3 = abjad.Container(r"\time 5/4 R1 * 5/4 \time 4/4 g''4")
        >>> container4 = abjad.Container(r"\time 6/8 R1 * 1/2")
        >>> auxjad.get.selection_is_full(container1[:])
        True
        >>> auxjad.get.selection_is_full(container2[:])
        True
        >>> auxjad.get.selection_is_full(container3[:])
        False
        >>> auxjad.get.selection_is_full(container4[:])
        False

    ..  error::

        If a selection is malformed, i.e. it has an underfilled measure before
        a time signature change, the function raises a :exc:`ValueError`
        exception. This is also the case when a selection starts in the middle
        of a measure.

        >>> container = abjad.Container(r"\time 5/4 g''1 \time 4/4 f'1")
        >>> auxjad.get.selection_is_full(container[:])
        ValueError: 'selection' is malformed, with an underfull measure
        preceding a time signature change

    ..  warning::

        The input selection must be a contiguous logical voice. When dealing
        with a container with multiple subcontainers (e.g. a score containing
        multiple staves), the best approach is to cycle through these
        subcontainers, applying this function to them individually.
    """
    if not isinstance(selection, abjad.Selection):
        raise TypeError("argument must be 'abjad.Selection'")
    if not selection.leaves().are_contiguous_logical_voice():
        raise ValueError("argument must be contiguous logical voice")
    return underfull_duration(selection) == abjad.Duration(0)



import abjad


def rests_to_multimeasure_rest(
    selection: abjad.Selection,
    *,
    ignore_clefs: bool = False,
    ignore_dynamics: bool = False,
) -> None:
    r"""Mutates an input |abjad.Selection| in place and has no return value;
    this function looks for measures filled with regular rests and converts
    them into an |abjad.MultimeasureRest|.

    Basic usage:
        Converts any measure filled with regular rests into a measure with a
        single multi-measure rest.

        >>> container = abjad.Container(r"\time 3/4 r2.")
        >>> abjad.show(container)

        ..  docs::

            {
                \time 3/4
                r2.
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-VIj5iWyLCG.png

        >>> auxjad.mutate.rests_to_multimeasure_rest(container[:])
        >>> abjad.show(container)

        ..  docs::

            {
                \time 3/4
                R1 * 3/4
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-uvebc7ft1zo.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.rests_to_multimeasure_rest(staff[:])
        >>> abjad.mutate.rests_to_multimeasure_rest(staff[:])

    Multiple rests:
        Works with measures with multiple regular rests.

        >>> container = abjad.Container(r"r2 r8.. r32 r16 r8 r16")
        >>> abjad.show(container)

        ..  docs::

            {
                r2
                r8..
                r32
                r16
                r8
                r16
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-jk4m1wzsyfa.png

        >>> auxjad.mutate.rests_to_multimeasure_rest(container[:])
        >>> abjad.show(container)

        ..  docs::

            {
                R1
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-z8u0cs3fzdi.png

    ..  note::

        When using |abjad.Container|'s, all time signatures in the output will
        be commented out with ``%%%.`` This is because Abjad only applies time
        signatures to containers that belong to a |abjad.Staff|. The present
        function works with either |abjad.Container| and |abjad.Staff|.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4")
        >>> abjad.show(container)

        ..  docs::

            {
                %%% \time 3/4 %%%
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-qtq55xbkkts.png

        >>> staff = abjad.Staff([container])
        >>> abjad.show(container)

        ..  docs::

            {
                \time 3/4
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-9hceg93vrmv.png

    Time signature changes:
        Works with selections from containers with multiple time signatures as
        well as notes.

        >>> container = abjad.Staff(
        ...     r"\time 3/4 r2. | "
        ...     "\time 6/8 r2. | "
        ...     "\time 5/4 c'1 ~ c'4 | r1 r4"
        ... )
        >>> abjad.show(container)

        ..  docs::

            \new Staff
            {
                \time 3/4
                r2.
                \time 6/8
                r2.
                \time 5/4
                c'1
                ~
                c'4
                r1
                r4
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-oQfFRihzEY.png

        >>> auxjad.mutate.rests_to_multimeasure_rest(container[:])
        >>> abjad.show(container)

        ..  docs::

            \new Staff
            {
                \time 3/4
                R1 * 3/4
                \time 6/8
                R1 * 3/4
                \time 5/4
                c'1
                ~
                c'4
                R1 * 5/4
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-a9sqdcznoq.png

    Tuplets:
        Works with containers with tuplets.

        >>> container = abjad.Container(r"\times 2/3 {r2 r2 r2}")
        >>> abjad.show(container)

        ..  docs::

            {
                \times 2/3
                {
                    r2
                    r2
                    r2
                }
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-480a9zqvk2a.png

        >>> auxjad.mutate.rests_to_multimeasure_rest(container[:])
        >>> abjad.show(container)

        ..  docs::

            {
                R1
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-r5yg3a3f97q.png

        It also works with containers with tuplets within tuplets.

        >>> container = abjad.Container(
        ...     r"r2 \times 2/3 {r2 r4} \times 4/5 {r2. \times 2/3 {r2 r4}}"
        ... )
        >>> abjad.show(container)

        ..  docs::

            {
                r2
                \times 2/3
                {
                    r2
                    r4
                }
                \times 4/5
                {
                    r2.
                    \times 2/3
                    {
                        r2
                        r4
                    }
                }
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-codydc205jw.png

        >>> auxjad.mutate.rests_to_multimeasure_rest(container[:])
        >>> abjad.show(container)

        ..  docs::

            {
                R1
                R1
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-f647t5j3jgw.png

    ``ignore_clefs``
        By default, the last clef of an empty measure is preserved when
        replacing it with a multi-measure rest:

        >>> staff = abjad.Staff(
        ...     r"\clef bass r4 r4 \times 2/3 {r4 r8} r4 "
        ...     r"\time 3/4 \clef treble r2. "
        ...     r"\time 5/4 r2 \clef bass r2."
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \clef "bass"
                r4
                r4
                \times 2/3
                {
                    r4
                    r8
                }
                r4
                \time 3/4
                \clef "treble"
                r2.
                \time 5/4
                r2
                \clef "bass"
                r2.
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-6GMRGmYkEQ.png

        >>> abjad.mutate.rests_to_multimeasure_rest(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \clef "bass"
                R1
                \time 3/4
                \clef "treble"
                R1 * 3/4
                \time 5/4
                \clef "bass"
                R1 * 5/4
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-UnL6ZoFoDC.png

        Invoke the mutation with ``ignore_clefs`` set to ``True`` to disable
        this behaviour and ignore all clefs:

        >>> staff = abjad.Staff(
        ...     r"\clef bass r4 r4 \times 2/3 {r4 r8} r4 "
        ...     r"\time 3/4 \clef treble r2. "
        ...     r"\time 5/4 r2 \clef bass r2."
        ... )
        >>> abjad.mutate.rests_to_multimeasure_rest(
        ...     staff[:],
        ...     ignore_clefs=True,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                R1
                \time 3/4
                R1 * 3/4
                \time 5/4
                R1 * 5/4
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-KGRZJ8fvQF.png

    ``ignore_dynamics``
        By default, the last dynamic or hairpin of an empty measure is
        preserved when replacing it with a multi-measure rest:

        >>> staff = abjad.Staff(r"c'1\p\< r2\! r2 d'1\f\> r2 r2\ppp")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                \<
                r2
                \!
                r2
                d'1
                \f
                \>
                r2
                r2
                \ppp
            }


        ..  figure:: ../_images/rests_to_multimeasure_rest-J9T5UY8r9w.png

        >>> abjad.mutate.rests_to_multimeasure_rest(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                \<
                R1
                \!
                d'1
                \f
                \>
                R1
                \ppp
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-77r9QeaZBA.png

        Invoke the mutation with ``ignore_dynamics`` set to ``True`` to disable
        this behaviour and ignore all dynamics and hairpins:

        >>> staff = abjad.Staff(r"c'1\p\< r2\! r2 d'1\f\> r2 r2\ppp")
        >>> abjad.mutate.rests_to_multimeasure_rest(
        ...     staff[:],
        ...     ignore_dynamics=True,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                \<
                R1
                d'1
                \f
                \>
                R1
            }

        ..  figure:: ../_images/rests_to_multimeasure_rest-3UhZukx9Pw.png

        ..  warning::

            Note that dynamics are only ignored when converting rests to
            multi-measure rests. All other dynamics are preserved in the score.
            This can result in problems displaying dynamics when one or more
            unterminated hairpins is present. In the example above, the last
            note's hairpin is unterminated and, because of that, LilyPond
            ignores all dynamics in that staff:

            >>> staff = abjad.Staff(r"c'1\p\< r2\! r2 d'1\f\> r2 r2\ppp")
            >>> abjad.mutate.rests_to_multimeasure_rest(
            ...     staff[:],
            ...     ignore_dynamics=True,
            ... )
            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'1
                \p
                \<
                R1
                d'1
                \f
                \>
                R1
            }

    ..  warning::

        The input selection must be a contiguous logical voice. When dealing
        with a container with multiple subcontainers (e.g. a score containing
        multiple staves), the best approach is to cycle through these
        subcontainers, applying this function to them individually.
    """
    if not isinstance(selection, abjad.Selection):
        raise TypeError("argument must be 'abjad.Selection'")
    if not selection.leaves().are_contiguous_logical_voice():
        raise ValueError("argument must be contiguous logical voice")
    if not isinstance(ignore_clefs, bool):
        raise TypeError("'ignore_clefs' must be 'bool'")
    if not isinstance(ignore_dynamics, bool):
        raise TypeError("'ignore_dynamics' must be 'bool'")

    measures = selection.group_by_measure()
    effective_time_signature = abjad.TimeSignature((4, 4))
    indicators_tuple = (
        abjad.BarLine,
        abjad.Fermata,
        abjad.KeySignature,
        abjad.LilyPondLiteral,
        abjad.MetronomeMark,
        abjad.RehearsalMark,
        abjad.Repeat,
        abjad.StaffChange,
        abjad.StartMarkup,
        abjad.StartTextSpan,
        abjad.StopTextSpan,
    )

    for measure in measures:
        head = abjad.select(measure).leaf(0)
        time_signature = abjad.get.indicator(head, abjad.TimeSignature)
        if time_signature is not None:
            effective_time_signature = time_signature
        if all([isinstance(leaf, abjad.Rest) for leaf in measure.leaves()]):
            if not ignore_clefs:
                for leaf in measure.leaves()[::-1]:
                    clef = abjad.get.indicator(leaf, abjad.Clef)
                    if clef is not None:
                        break
            if not ignore_dynamics:
                for leaf in measure.leaves()[::-1]:
                    dynamics = abjad.get.indicator(
                        leaf,
                        (
                            abjad.Dynamic,
                            abjad.StartHairpin,
                            abjad.StopHairpin,
                        ),
                    )
                    if dynamics is not None:
                        break
            duration = abjad.get.duration(measure)
            if duration == effective_time_signature.duration:
                if duration == 1:
                    multiplier = None
                else:
                    multiplier = abjad.Multiplier(duration)
                multimeasure_rest = abjad.MultimeasureRest(
                    (4, 4),
                    multiplier=multiplier,
                )
                if time_signature is not None:
                    abjad.attach(time_signature, multimeasure_rest)
                if not ignore_clefs and clef is not None:
                    abjad.attach(clef, multimeasure_rest)
                if not ignore_dynamics and dynamics is not None:
                    abjad.attach(dynamics, multimeasure_rest)
                for indicator in abjad.get.indicators(head):
                    if isinstance(indicator, indicators_tuple):
                        abjad.attach(indicator, multimeasure_rest)
                abjad.mutate.replace(measure, multimeasure_rest)



import random
from typing import Optional, Union

import abjad

from .. import get, mutate


class Hocketer:
    r"""A hocket generator that takes an |abjad.Container| (or child class) as
    input and randomly distributes its logical ties among different staves.

    Basic usage:
        Calling the object will return a :obj:`tuple` of |abjad.Selection|
        generated by the hocket process and ready to be assigned to an
        |abjad.Score|. Each call will generate a new random hocket from the
        input container.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> abjad.show(container)

        ..  docs::

            {
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Hocketer-svcW7YQBUu.png

        >>> hocketer = auxjad.Hocketer(container)
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    r2
                    e'4
                    r4
                }
                \new Staff
                {
                    c'4
                    d'4
                    r4
                    f'4
                }
            >>

        ..  figure:: ../_images/Hocketer-dh5lutbv09v.png

        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    c'4
                    d'4
                    e'4
                    r4
                }
                \new Staff
                {
                    r2.
                    f'4
                }
            >>

        ..  figure:: ../_images/Hocketer-gj278dv7vff.png

    Indexing:
        Alternatively, it is possible to retrieve an |abjad.Selection| for
        each individual voice generated by the process by indexing or slicing
        the the object itself. In the case below, the hocket process is invoked
        in the third line, and the individual selections are retrieved in the
        loop.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> hocketer = auxjad.Hocketer(container, n_voices=5, k=3)
        >>> hocketer()
        >>> score = abjad.Score()
        >>> for selection in hocketer[:]:
        ...     staff = abjad.Staff(selection)
        ...     score.append(staff)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    r2.
                    f'4
                }
                \new Staff
                {
                    c'4
                    r4
                    e'4
                    r4
                }
                \new Staff
                {
                    r4
                    d'4
                    e'4
                    r4
                }
                \new Staff
                {
                    c'4
                    d'4
                    r4
                    f'4
                }
                \new Staff
                {
                    c'4
                    r4
                    e'4
                    r4
                }
            >>

        ..  figure:: ../_images/Hocketer-5r3ybkxhmd4.png

        Individual staves can also be retrieved as shown below.

        >>> staff = abjad.Staff(hocketer[0])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                r4
                e'4
                f'4
            }

        ..  figure:: ../_images/Hocketer-sodb0btsuhq.png

        >>> partial_score = abjad.Score()
        >>> for selection in hocketer[1:4]:
        ...     staff = abjad.Staff(selection)
        ...     partial_score.append(selection)
        >>> abjad.show(partial_score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    c'4
                    d'4
                    r2
                }
                \new Staff
                {
                    r4
                    d'4
                    e'4
                    r4
                }
                \new Staff
                {
                    R1
                }
            >>

        ..  figure:: ../_images/Hocketer-3gsh3qtl21d.png

    :attr:`current_window`:
        To get the result of the last operation, use the property
        :attr:`current_window`.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> hocketer = auxjad.Hocketer(container)
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    c'4
                    r4
                    e'4
                    r4
                }
                \new Staff
                {
                    r4
                    d'4
                    r4
                    f'4
                }
            >>

        ..  figure:: ../_images/Hocketer-i2jt38x28fd.png

        >>> music = hocketer.current_window
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    c'4
                    r4
                    e'4
                    r4
                }
                \new Staff
                {
                    r4
                    d'4
                    r4
                    f'4
                }
            >>

        ..  figure:: ../_images/Hocketer-4291u773dng.png

    Arguments and properties:
        This class has many keyword arguments, all of which can be altered
        after instantiation using properties with the same names as shown
        below. :attr:`weights` set the individual weight of a given voice (must
        be a :obj:`list` of length equal to :attr:`n_voices`). :attr:`k`
        defines the number of times that the process is applied to each logical
        tie. Setting :attr:`force_k_voices` to ``True`` ensure that a single
        logical tie is distributed to exactly :attr:`k` voices. Setting
        :attr:`explode_chords` to ``True`` will distribute individual pitches
        from chords into unique voices. :attr:`pitch_ranges` defines the pitch
        range of each voice; it takes a list of |abjad.PitchRange|'s.
        :attr:`disable_rewrite_meter` disables the
        |abjad.Meter.rewrite_meter()| mutation which is applied to the
        container after every call. Any measure filled with rests will be
        rewritten using a multi-measure rest; set the
        :attr:`use_multimeasure_rests` to ``False`` to disable this behaviour.
        The properties :attr:`boundary_depth`, :attr:`maximum_dot_count`, and
        :attr:`rewrite_tuplets` are passed as arguments to
        |abjad.Meter.rewrite_meter()|, see its documentation for more
        information. Setting the property :attr:`omit_time_signatures` to
        ``True`` will remove all time signatures from the output (``False`` by
        default).

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4 | f'4 g'4 a'4")
        >>> hocketer = auxjad.Hocketer(container,
        ...                            n_voices=3,
        ...                            weights=[1, 2, 5],
        ...                            k=2,
        ...                            force_k_voices=True,
        ...                            explode_chords=True,
        ...                            pitch_ranges=[
        ...                                abjad.PitchRange("[C4, D6]"),
        ...                                abjad.PitchRange("[C2, A4]"),
        ...                                abjad.PitchRange("[C1, E3]"),
        ...                            ],
        ...                            disable_rewrite_meter=True,
        ...                            use_multimeasure_rests=False,
        ...                            omit_time_signatures=True,
        ...                            boundary_depth=0,
        ...                            maximum_dot_count=1,
        ...                            rewrite_tuplets=False,
        ...                            )
        >>> hocketer.n_voices
        3
        >>> hocketer.weights
        [1, 2, 5]
        >>> hocketer.k
        2
        >>> hocketer.force_k_voices
        True
        >>> hocketer.explode_chords
        True
        >>> hocketer.pitch_ranges
        [PitchRange('[C4, D6]'), PitchRange('[C2, A4]'),
        PitchRange('[C1, E3]')]
        >>> hocketer.disable_rewrite_meter
        True
        >>> not hocketer.use_multimeasure_rests
        False
        >>> hocketer.omit_time_signatures
        True
        >>> hocketer.boundary_depth
        0
        >>> hocketer.maximum_dot_count
        1
        >>> hocketer.rewrite_tuplets
        False

        Use the properties below to change these values after initialisation.

        >>> hocketer.n_voices = 5
        >>> hocketer.weights = [1, 1, 1, 2, 7]
        >>> hocketer.k = 3
        >>> hocketer.force_k_voices = False
        >>> hocketer.explode_chords = False
        >>> hocketer.pitch_ranges = None
        >>> hocketer.disable_rewrite_meter = False
        >>> hocketer.use_multimeasure_rests = True
        >>> hocketer.omit_time_signatures = False
        >>> hocketer.boundary_depth = 1
        >>> hocketer.maximum_dot_count = 2
        >>> hocketer.rewrite_tuplets = True
        >>> hocketer.n_voices
        5
        >>> hocketer.weights
        [1, 1, 1, 2, 7]
        >>> hocketer.k
        3
        >>> not hocketer.force_k_voices
        False
        >>> hocketer.explode_chords
        False
        >>> hocketer.pitch_ranges
        None
        >>> not hocketer.disable_rewrite_meter
        False
        >>> hocketer.use_multimeasure_rests
        True
        >>> hocketer.omit_time_signatures
        False
        >>> hocketer.boundary_depth
        1
        >>> hocketer.maximum_dot_count
        2
        >>> hocketer.rewrite_tuplets
        True

    :attr:`n_voices`:
        Use the optional argument :attr:`n_voices` to set the number of
        different staves in the output (default is 2).

        >>> container = abjad.Container(r"c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")
        >>> hocketer = auxjad.Hocketer(container, n_voices=3)
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    c'8
                    r8
                    e'8
                    r8
                    g'8
                    r4.
                }
                \new Staff
                {
                    r8
                    d'8
                    r4
                    r8
                    a'8
                    r4
                }
                \new Staff
                {
                    r4.
                    f'8
                    r4
                    b'8
                    c''8
                }
            >>

        ..  figure:: ../_images/Hocketer-qf9niqf7em.png

    :func:`len()`:
        Applying the :func:`len()` function to the hocketer will return the
        current number of voices to be output by the hocketer.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4 ~ | f'2 g'2")
        >>> hocketer = auxjad.Hocketer(container, n_voices=7)
        >>> len(hocketer)
        6

    :attr:`pitch_ranges`:
        The property :attr:`pitch_ranges` defines the pitch range of each
        voice. It takes a list of |abjad.PitchRange|'s.


        >>> container = abjad.Container(r"c' d' e f' g a' b' c''")
        >>> hocketer = auxjad.Hocketer(container,
        ...                            n_voices=3,
        ...                            pitch_ranges=[
        ...                                abjad.PitchRange("[C4, C5]"),
        ...                                abjad.PitchRange("[C4, C5]"),
        ...                                abjad.PitchRange("[C3, B3]"),
        ...                            ],
        ...                            )
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.attach(abjad.Clef('bass'), abjad.select(score[2]).leaf(0))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    c'4
                    d'4
                    r4
                    f'4
                    r4
                    a'4
                    r2
                }
                \new Staff
                {
                    R1
                    r2
                    b'4
                    c''4
                }
                \new Staff
                {
                    \clef "bass"
                    r2
                    e4
                    r4
                    g4
                    r2.
                }
            >>

        ..  figure:: ../_images/Hocketer-akuL9BDgw8.png

        Note that changing :attr:`n_voices` will reset :attr:`pitch_ranges`:

        >>> container = abjad.Container(
        ...     r"\time 3/4 c'4 d'4 e'4 \time 2/4 f'4 g'4"
        ... )
        >>> hocketer = auxjad.Hocketer(container,
        ...                            n_voices=3,
        ...                            pitch_ranges=[
        ...                                abjad.PitchRange("[C4, D6]"),
        ...                                abjad.PitchRange("[C2, A4]"),
        ...                                abjad.PitchRange("[C1, E3]"),
        ...                            ],
        ...                            )
        >>> hocketer.pitch_ranges
        [PitchRange("[C4, D6]"), PitchRange("[C2, A4]"),
        PitchRange("[C1, E3]")]         ]
        >>> hocketer.n_voices = 4
        >>> hocketer.pitch_ranges
        None

    :attr:`weights`:
        Set :attr:`weights` to a :obj:`list` of numbers (either :obj:`float` or
        :obj:`int`) to give different weights to each voice. By default, all
        voices have equal weight. The :obj:`list` in :attr:`weights` must have
        the same length as the number of voices. In the example below,
        :attr:`weights` is a :obj:`list` of length ``2`` (matching the default
        two voices). The second voice has a higher weight to it, and receives
        more notes from the hocket process as expected.

        >>> container = abjad.Container(r"c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")
        >>> hocketer = auxjad.Hocketer(container,
        ...                            weights=[2.1, 5.7],
        ...                            )
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    r8
                    d'8
                    r4
                    g'8
                    r4.
                }
                \new Staff
                {
                    c'8
                    r8
                    e'8
                    f'8
                    r8
                    a'8
                    b'8
                    c''8
                }
            >>

        ..  figure:: ../_images/Hocketer-xyecbupzmm.png

    :meth:`reset_weights`:
        Use the method :meth:`reset_weights` to reset the weights back to their
        default values.

        >>> container = abjad.Container(r"c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")
        >>> hocketer = auxjad.Hocketer(container,
        ...                            weights=[2.1, 5.7],
        ...                            )
        >>> hocketer.weights
        [2.1, 5.7]
        >>> hocketer.reset_weights()
        >>> hocketer.weights
        [1.0, 1.0]

    :attr:`k`:
        The argument :attr:`k` is an :obj:`int` defining the number of times
        that the process is applied to each logical tie. By default, :attr:`k`
        is set to ``1``, so each logical tie is assigned to a single voice.
        Changing this to a higher value will increase the chance of a logical
        tie appearing for up to :attr:`k` different voices.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> hocketer = auxjad.Hocketer(container, n_voices=4, k=2)
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    c'4
                    d'4
                    e'4
                    r4
                }
                \new Staff
                {
                    r2
                    e'4
                    f'4
                }
                \new Staff
                {
                    r2.
                    f'4
                }
                \new Staff
                {
                    r4
                    d'4
                    r2
                }
            >>

        ..  figure:: ../_images/Hocketer-ytjqsp7r2bn.png

    :attr:`force_k_voices`:
        It is important to note that changing :attr:`k` to a higher value does
        not guarantee each logical tie will appear in :attr:`k` different
        voices. By default, :attr:`k` only defines how many times each logical
        tie is processed by the hocket process, which may select the same voice
        more than once. To ensure that each logical tie appears in :attr:`k`
        unique voices, set the optional keyword argument :attr:`force_k_voices`
        to ``True`` as shown below.

        >>> container = abjad.Container(r"c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")
        >>> hocketer = auxjad.Hocketer(container,
        ...                            n_voices=3,
        ...                            k=2,
        ...                            force_k_voices=True,
        ...                            )
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    c'8
                    d'8
                    r8
                    f'8
                    g'8
                    a'8
                    b'8
                    c''8
                }
                \new Staff
                {
                    c'8
                    r8
                    e'8
                    f'8
                    r4
                    b'8
                    r8
                }
                \new Staff
                {
                    r8
                    d'8
                    e'8
                    r8
                    g'8
                    a'8
                    r8
                    c''8
                }
            >>

        ..  figure:: ../_images/Hocketer-9limlogk0b8.png

    ..  error::

        Setting :attr:`force_k_voices` to ``True`` when :attr:`k` is larger
        than :attr:`n_voices` will raise a :exc:`ValueError` exception:

        >>> container = abjad.Container(r"c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")
        >>> hocketer = auxjad.Hocketer(container, n_voices=4, k=5)
        >>> hocketer.force_k_voices = True
        ValueError: 'force_k_voices' cannot be set to True if 'k' > 'n_voices',
        change 'k' first

    :attr:`explode_chords`:
        If :attr:`explode_chords` is set to ``True``, chords will not be
        considered as a single leaf to be distributed but rather as a
        collection of individual pitches, which are then distributed among the
        voices. Compare:

        >>> container = abjad.Container(
        ...     r"<c' e' g'>4 <d' f' a'>4 <e' g' b'>4 <f' a' c'>4"
        ... )
        >>> hocketer = auxjad.Hocketer(container,
        ...                            n_voices=3,
        ...                            )
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    r2
                    <e' g' b'>4
                    r4
                }
                \new Staff
                {
                    <c' e' g'>4
                    <d' f' a'>4
                    r2
                }
                \new Staff
                {
                    r2.
                    <c' f' a'>4
                }
            >>

        ..  figure:: ../_images/Hocketer-6gwHXAs6IY.png

        >>> container = abjad.Container(
        ...     r"<c' e' g'>4 <d' f' a'>4 <e' g' b'>4 <f' a' c'>4"
        ... )
        >>> hocketer = auxjad.Hocketer(container,
        ...                            n_voices=3,
        ...                            explode_chords=True,
        ...                            )
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    e'4
                    f'4
                    e'4
                    a'4
                }
                \new Staff
                {
                    g'4
                    d'4
                    b'4
                    c'4
                }
                \new Staff
                {
                    c'4
                    a'4
                    g'4
                    f'4
                }
            >>

        ..  figure:: ../_images/Hocketer-gGsBUch6Ai.png

    ..  note::

        It is very important to note that :attr:`explode_chords` does not take
        :attr:`weights` nor :attr:`k` in consideration. It does, however, take
        :attr:`pitch_ranges` into account:

        >>> container = abjad.Container(
        ...     r"<c' e' g'>4 <d' f' a'>4 <e' g' b'>4 <f' a' c'>4"
        ... )
        >>> hocketer = auxjad.Hocketer(container,
        ...                            n_voices=3,
        ...                            explode_chords=True,
        ...                            pitch_ranges=[
        ...                                abjad.PitchRange("[E4, C5]"),
        ...                                abjad.PitchRange("[E4, C5]"),
        ...                                abjad.PitchRange("[C4, F4]"),
        ...                            ],
        ...                            )
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    e'4
                    f'4
                    g'4
                    c''4
                }
                \new Staff
                {
                    g'4
                    a'4
                    b'4
                    a'4
                }
                \new Staff
                {
                    c'4
                    d'4
                    e'4
                    f'4
                }
            >>

        ..  figure:: ../_images/Hocketer-Rycx76se89.png

    ..  warning::

        When setting :attr:`explode_chords` to ``True``, if :attr:`n_voices` is
        less than the number of pitches in a chord then only the first
        :attr:`n_voices`-pitches will be considered in the distribution.

    :attr:`disable_rewrite_meter`:
        By default, this class uses the |abjad.Meter.rewrite_meter()|
        mutation, which is necessary for cleaning up the rests.

        >>> container = abjad.Container(r"c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")
        >>> hocketer = auxjad.Hocketer(container)
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    c'8
                    d'8
                    r4
                    r8
                    a'8
                    r4
                }
                \new Staff
                {
                    r4
                    e'8
                    f'8
                    g'8
                    r8
                    b'8
                    c''8
                }
            >>

        ..  figure:: ../_images/Hocketer-qtuvb2zrlz.png

        Set :attr:`disable_rewrite_meter` to ``True`` in order to disable this
        behaviour.

        >>> container = abjad.Container(r"c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")
        >>> hocketer = auxjad.Hocketer(container,
        ...                            disable_rewrite_meter=True,
        ...                            )
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    c'8
                    d'8
                    r8
                    r8
                    r8
                    a'8
                    r8
                    r8
                }
                \new Staff
                {
                    r8
                    r8
                    e'8
                    f'8
                    g'8
                    r8
                    b'8
                    c''8
                }
            >>

        ..  figure:: ../_images/Hocketer-l8cn8w7cju.png

    :attr:`use_multimeasure_rests`:
        By default, this class rewrites all measures that are filled with
        rests, replacing the rests by a multi-measure rest.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4 f'4 g'4 a'4")
        >>> hocketer = auxjad.Hocketer(container)
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \time 3/4
                    R1 * 3/4
                    r4
                    g'4
                    a'4
                }
                \new Staff
                {
                    \time 3/4
                    c'4
                    d'4
                    e'4
                    f'4
                    r2
                }
            >>

        ..  figure:: ../_images/Hocketer-ri1ursa6esi.png

        Set :attr:`use_multimeasure_rests` to ``False`` to disable this
        behaviour.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4 f'4 g'4 a'4")
        >>> hocketer = auxjad.Hocketer(container,
        ...                            use_multimeasure_rests=False,
        ...                            )
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \time 3/4
                    r2.
                    r4
                    g'4
                    a'4
                }
                \new Staff
                {
                    \time 3/4
                    c'4
                    d'4
                    e'4
                    f'4
                    r2
                }
            >>

        ..  figure:: ../_images/Hocketer-20idpp518lf.png

    :attr:`contents`:
        Use the property :attr:`contents` to get the input container upon which
        the hocketer operates. Notice that :attr:`contents` remains invariant
        after any shuffling or rotation operations (use :attr:`current_window`
        for the transformed selection of music). :attr:`contents` can be used
        to change the |abjad.Container| to be shuffled.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> hocketer = auxjad.Hocketer(container)
        >>> abjad.show(hocketer.contents)

        ..  docs::

            {
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Hocketer-t07z297s2e.png

        >>> hocketer()
        >>> abjad.show(hocketer.contents)

        ..  docs::

            {
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Hocketer-7g6cbvfgr1j.png

        >>> hocketer.contents = abjad.Container(r"cs2 ds2")
        >>> abjad.show(hocketer.contents)

        ..  docs::

            {
                cs2
                ds2
            }

        ..  figure:: ../_images/Hocketer-2l7eywciib3.png

    Tweaking |abjad.Meter.rewrite_meter()|:
        This function uses the default logical tie splitting algorithm from
        |abjad.Meter.rewrite_meter()|.

        >>> container = abjad.Container(r"c'4. d'8 e'2")
        >>> hocketer = auxjad.Hocketer(container,
        ...                            n_voices=1,
        ...                            )
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    c'4.
                    d'8
                    e'2
                }
            >>

        ..  figure:: ../_images/Hocketer-g3fmvrkd6nw.png

        Set :attr:`boundary_depth` to a different number to change its
        behaviour.

        >>> hocketer = auxjad.Hocketer(container,
        ...                            n_voices=1,
        ...                            boundary_depth=1,
        ...                            )
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(staff)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    c'4
                    ~
                    c'8
                    d'8
                    e'2
                }
            >>

        ..  figure:: ../_images/Hocketer-g7r2oxsx8zc.png

        Other arguments available for tweaking the output of
        |abjad.Meter.rewrite_meter()| are :attr:`maximum_dot_count` and
        :attr:`rewrite_tuplets`, which work exactly as the identically named
        arguments of |abjad.Meter.rewrite_meter()|.

        This class also accepts the arguments ``fuse_across_groups_of_beats``,
        ``fuse_quadruple_meter``, ``fuse_triple_meter``, and
        ``extract_trivial_tuplets``, which are passed on to
        |auxjad.mutate.prettify_rewrite_meter()| (the latter can be disabled
        by setting ``prettify_rewrite_meter`` to ``False``). See the
        documentation of this function for more details on these arguments.

    Time signature changes and nested tuplets:
        This class can handle time signature changes as well as nested tuplets.

        >>> container = abjad.Container(
        ...     r"\time 5/4 r4 \times 2/3 {c'4 d'2} e'4. f'8 "
        ...     r"\times 4/5 {\time 4/4 g'2. \times 2/3 {a'8 r8 b'2}}"
        ... )
        >>> hocketer = auxjad.Hocketer(container,
        ...                            n_voices=4,
        ...                            k=2,
        ...                            )
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \time 5/4
                    r4
                    \times 2/3
                    {
                        c'4
                        d'2
                    }
                    r4.
                    f'8
                    \times 4/5
                    {
                        \time 4/4
                        r2.
                        \times 2/3
                        {
                            a'8
                            r8
                            r2
                        }
                    }
                }
                \new Staff
                {
                    \time 5/4
                    R1 * 5/4
                    \time 4/4
                    R1
                }
                \new Staff
                {
                    \time 5/4
                    r4
                    \times 2/3
                    {
                        r4
                        d'2
                    }
                    e'4.
                    f'8
                    \times 4/5
                    {
                        \time 4/4
                        r2.
                        \times 2/3
                        {
                            a'8
                            r8
                            b'2
                        }
                    }
                }
                \new Staff
                {
                    \time 5/4
                    r4
                    \times 2/3
                    {
                        c'4
                        r2
                    }
                    r2
                    \times 4/5
                    {
                        \time 4/4
                        g'2.
                        \times 2/3
                        {
                            r4
                            b'2
                        }
                    }
                }
            >>

        ..  figure:: ../_images/Hocketer-seuga1yah9h.png

    :attr:`omit_time_signatures`:
        To disable time signatures altogether, initialise this class with the
        keyword argument :attr:`omit_time_signatures` set to ``True`` (default
        is ``False``), or use the :attr:`omit_time_signatures` property after
        initialisation. It is recommended to also set
        :attr:`use_multimeasure_rests` to ``False``, as those are created
        according to the original time signatures.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4 f'4 g'4 a'4")
        >>> hocketer = auxjad.Hocketer(container,
        ...                            omit_time_signatures=True,
        ...                            use_multimeasure_rests=False,
        ...                            )
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    r2.
                    r4
                    g'4
                    a'4
                }
                \new Staff
                {
                    c'4
                    d'4
                    e'4
                    f'4
                    r2
                }
            >>

        ..  figure:: ../_images/Hocketer-6w9r9iicp29.png

    Indicators:
        Dynamics and hairpins are supported.

        >>> container = abjad.Container(
        ...     r"c'2-.\p\< d'2-.\f\> e'1 f'2.\pp\< g'4--\p "
        ...     r"a'2\ff\> b'2\p\> ~ b'2 c''2\ppp"
        ... )
        >>> hocketer = auxjad.Hocketer(container,
        ...                            n_voices=3,
        ...                            k=2,
        ...                            force_k_voices=True,
        ...                            )
        >>> music = hocketer()
        >>> score = abjad.Score()
        >>> for selection in music:
        ...     score.append(abjad.Staff(selection))
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    c'2
                    \p
                    - \staccato
                    \<
                    d'2
                    \f
                    - \staccato
                    \>
                    e'1
                    f'2.
                    \pp
                    \<
                    g'4
                    \p
                    - \tenuto
                    a'2
                    \ff
                    \>
                    r2
                    \p
                    r2
                    c''2
                    \ppp
                }
                \new Staff
                {
                    r2
                    d'2
                    \f
                    - \staccato
                    \>
                    e'1
                    r2.
                    \pp
                    g'4
                    \p
                    - \tenuto
                    r2
                    b'2
                    \>
                    ~
                    b'2
                    r2
                    \ppp
                }
                \new Staff
                {
                    c'2
                    \p
                    - \staccato
                    \<
                    r2
                    \f
                    R1
                    f'2.
                    \pp
                    \<
                    r4
                    \p
                    a'2
                    \ff
                    \>
                    b'2
                    \p
                    \>
                    ~
                    b'2
                    c''2
                    \ppp
                }
            >>

        ..  figure:: ../_images/Hocketer-zel37ga6xob.png

    ..  tip::

        The functions |auxjad.mutate.remove_repeated_dynamics()| and
        |auxjad.mutate.reposition_clefs()| can be used to clean the output
        and remove repeated dynamics and unnecessary clef changes.

    ..  warning::

        Do note that some elements that span multiple notes (such as ottava
        indicators, manual beams, etc.) can become problematic when notes
        containing them are split into two. As a rule of thumb, it is always
        better to attach those to the music after the fading process has ended.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_contents",
        "_n_voices",
        "_weights",
        "_k",
        "_pitch_ranges",
        "_force_k_voices",
        "_explode_chords",
        "_disable_rewrite_meter",
        "_use_multimeasure_rests",
        "_voices",
        "_time_signatures",
        "_boundary_depth",
        "_maximum_dot_count",
        "_rewrite_tuplets",
        "_omit_time_signatures",
        "_prettify_rewrite_meter",
        "_extract_trivial_tuplets",
        "_fuse_across_groups_of_beats",
        "_fuse_quadruple_meter",
        "_fuse_triple_meter",
    )

    ### INITIALISER ###

    def __init__(
        self,
        contents: abjad.Container,
        *,
        n_voices: int = 2,
        pitch_ranges: Optional[list] = None,
        weights: Optional[list] = None,
        k: int = 1,
        force_k_voices: bool = False,
        explode_chords: bool = False,
        disable_rewrite_meter: bool = False,
        use_multimeasure_rests: bool = True,
        omit_time_signatures: bool = False,
        boundary_depth: Optional[int] = None,
        maximum_dot_count: Optional[int] = None,
        rewrite_tuplets: bool = True,
        prettify_rewrite_meter: bool = True,
        extract_trivial_tuplets: bool = True,
        fuse_across_groups_of_beats: bool = True,
        fuse_quadruple_meter: bool = True,
        fuse_triple_meter: bool = True,
    ) -> None:
        self.contents = contents
        self._voices = None
        self._n_voices = n_voices
        self._k = k
        if weights is not None:
            self.weights = weights
        else:
            self.reset_weights()
        self.pitch_ranges = pitch_ranges
        self.force_k_voices = force_k_voices
        self.explode_chords = explode_chords
        self.disable_rewrite_meter = disable_rewrite_meter
        self.use_multimeasure_rests = use_multimeasure_rests
        self.omit_time_signatures = omit_time_signatures
        self.boundary_depth = boundary_depth
        self.maximum_dot_count = maximum_dot_count
        self.rewrite_tuplets = rewrite_tuplets
        self.prettify_rewrite_meter = prettify_rewrite_meter
        self.extract_trivial_tuplets = extract_trivial_tuplets
        self.fuse_across_groups_of_beats = fuse_across_groups_of_beats
        self.fuse_quadruple_meter = fuse_quadruple_meter
        self.fuse_triple_meter = fuse_triple_meter

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r"""Returns interpreter representation of :attr:`contents`."""
        return abjad.lilypond(self._contents)

    def __len__(self) -> int:
        r"""Returns the number of voices of the hocketer."""
        return self._n_voices

    def __call__(self) -> tuple[abjad.Selection]:
        r"""Calls the hocket process, returning a :obj:`tuple` of
        |abjad.Selection|.
        """
        self._make_music()
        return self.current_window

    def __getitem__(self, key: int) -> abjad.Selection:
        r"""Returns one or more voices of the output of the hocketer through
        indexing or slicing.
        """
        output = []
        for voice in self._voices[key]:
            voice_ = abjad.mutate.copy(voice)
            output.append(voice_)
        return tuple(output)

    ### PUBLIC METHODS ###

    def reset_weights(self) -> None:
        r"""Resets the weight vector of all voices to an uniform
        distribution.
        """
        self._weights = [1.0 for _ in range(self.__len__())]

    ### PRIVATE METHODS ###

    def _make_music(self) -> None:
        r"""Runs the hocket process, returning a :obj:`tuple` of
        |abjad.Container()|. It distributes the logical ties from the
        :attr:`contents` into different voices. Voices can have different
        weights, and the process of distributing a same logical tie can be run
        more than once (defined by the attribute :attr:`k`).
        """
        # distributing logical ties into voices
        dummy_voices = self._hocket_process()
        # handling dynamics and slurs
        for voice in dummy_voices:
            mutate.reposition_dynamics(voice[:])
            mutate.reposition_slurs(voice[:])
        # rewriting meter
        if not self._disable_rewrite_meter:
            for voice in dummy_voices:
                mutate.auto_rewrite_meter(
                    voice,
                    meter_list=self._time_signatures,
                    boundary_depth=self._boundary_depth,
                    maximum_dot_count=self._maximum_dot_count,
                    rewrite_tuplets=self._rewrite_tuplets,
                    prettify_rewrite_meter=self._prettify_rewrite_meter,
                    extract_trivial_tuplets=self._extract_trivial_tuplets,
                    fuse_across_groups_of_beats=(self.fuse_across_groups_of_beats),
                    fuse_quadruple_meter=self._fuse_quadruple_meter,
                    fuse_triple_meter=self._fuse_triple_meter,
                )
        # handling empty tuplets and multi-measure rests
        if self._use_multimeasure_rests:
            for voice in dummy_voices:
                mutate.rests_to_multimeasure_rest(voice[:])
        # output
        self._voices = []
        for voice in dummy_voices:
            self._voices.append(voice[:])
            voice[:] = []

    def _hocket_process(self) -> abjad.Container:
        r"""Replaces notes and chords for silences if voice not in the selected
        :obj:`list` for a given logical tie.
        """
        indicators_tuple = (
            abjad.BarLine,
            abjad.Clef,
            abjad.Dynamic,
            abjad.Fermata,
            abjad.KeySignature,
            abjad.LilyPondLiteral,
            abjad.MetronomeMark,
            abjad.Ottava,
            abjad.RehearsalMark,
            abjad.Repeat,
            abjad.StaffChange,
            abjad.StartHairpin,
            abjad.StartMarkup,
            abjad.StartPhrasingSlur,
            abjad.StartSlur,
            abjad.StartTextSpan,
            abjad.StopHairpin,
            abjad.StopPhrasingSlur,
            abjad.StopSlur,
            abjad.StopTextSpan,
            abjad.TimeSignature,
        )
        dummy_voices = [abjad.mutate.copy(self._contents) for _ in range(self._n_voices)]
        selected_voices = self._select_voices()
        for voice_index, voice in enumerate(dummy_voices):
            logical_ties = abjad.select(voice).logical_ties()
            for logical_tie, selected_indeces in zip(
                logical_ties,
                selected_voices,
            ):
                if voice_index not in selected_indeces:
                    for leaf in logical_tie:
                        rest = abjad.Rest(leaf.written_duration)
                        for indicator in abjad.get.indicators(leaf):
                            if isinstance(indicator, indicators_tuple):
                                abjad.attach(indicator, rest)
                        abjad.mutate.replace(leaf, rest)
                elif isinstance(logical_tie.head, abjad.Chord) and self._explode_chords:
                    pitch_number = selected_indeces.index(voice_index)
                    pitch = logical_tie.head.written_pitches[pitch_number]
                    for leaf in logical_tie:
                        note = abjad.Note(pitch, leaf.written_duration)
                        for indicator in abjad.get.indicators(leaf):
                            abjad.attach(indicator, note)
                        abjad.mutate.replace(leaf, note)
        return dummy_voices

    def _select_voices(self) -> list[int]:
        r"""Creates a :obj:`list` of selected voices for each logical tie."""
        selected_voices = []
        if not self._force_k_voices:
            for logical_tie in abjad.select(self._contents).logical_ties():
                # 1st case: exploding chords and ignoring k
                if isinstance(logical_tie.head, abjad.Chord) and self._explode_chords:
                    pitches = logical_tie.head.written_pitches
                    counter = 0
                    while True:
                        if len(pitches) > self._n_voices:
                            sample_k = self._n_voices
                        else:
                            sample_k = len(pitches)
                        voices = random.sample(
                            list(range(self._n_voices)),
                            k=sample_k,
                        )
                        counter += 1
                        if all(
                            self._pitch_in_range(pitch, voice)
                            for voice, pitch in zip(voices, pitches)
                        ):
                            break
                        elif counter >= 1000:
                            raise RuntimeError(
                                "No good distribution of chord "
                                "found, please check the pitch "
                                "ranges or try another seed."
                            )
                    selected_voices.append(voices)
                # 2nd case: distributing leaves into up to k voices, though
                # not enforcing k voices
                else:
                    pitch = self._get_pitch_from_logical_tie(logical_tie)
                    counter = 0
                    while True:
                        voices = random.choices(
                            list(range(self._n_voices)),
                            weights=self._weights,
                            k=self._k,
                        )
                        counter += 1
                        if all(self._pitch_in_range(pitch, voice) for voice in voices):
                            break
                        if counter >= 1000:
                            raise RuntimeError(
                                "No good distribution of notes "
                                "found, please check pitch "
                                "ranges or try another seed."
                            )
                    selected_voices.append(voices)
        # 3rd case: distributing leaves into exactly k voices
        else:
            for logical_tie in abjad.select(self._contents).logical_ties():
                pitch = self._get_pitch_from_logical_tie(logical_tie)
                counter = 0
                voices = []
                while len(voices) < self._k:
                    voice = random.choices(
                        list(range(self._n_voices)),
                        weights=self._weights,
                    )[0]
                    counter += 1
                    if voice in voices:
                        continue
                    if self._pitch_in_range(pitch, voice):
                        voices.append(voice)
                    if counter >= 1000:
                        raise RuntimeError(
                            "No good distribution of notes "
                            "found, please check pitch "
                            "ranges or try another seed."
                        )
                selected_voices.append(voices)
        return selected_voices

    def _pitch_in_range(
        self,
        pitch: Union[abjad.Pitch, abjad.PitchSegment, None],
        voice: int,
    ) -> bool:
        r"""Checks if a pitch is playable by a specific voice."""
        if self._pitch_ranges is None or pitch is None:
            return True
        elif isinstance(pitch, abjad.PitchSegment):
            if min(pitch) in self._pitch_ranges[voice] and max(pitch) in self._pitch_ranges[voice]:
                return True
        if pitch in self._pitch_ranges[voice]:
            return True
        return False

    def _get_lilypond_format(self) -> str:
        r"""Returns interpreter representation of  :attr:`contents`."""
        return self.__repr__()

    @staticmethod
    def _get_pitch_from_logical_tie(
        logical_tie: abjad.LogicalTie,
    ) -> Union[
        abjad.Pitch,
        abjad.PitchSegment,
        None,
    ]:
        r"""Gets the pitch of the first leaf of a logical tie."""
        if isinstance(logical_tie.head, abjad.Note):
            return logical_tie.head.written_pitch
        elif isinstance(logical_tie.head, abjad.Chord):
            return logical_tie.head.written_pitches
        else:
            return None

    @staticmethod
    def _remove_all_time_signatures(
        container: abjad.Container,
    ) -> None:
        r"""Removes all time signatures of an |abjad.Container|."""
        for leaf in abjad.select(container).leaves():
            if abjad.get.effective(leaf, abjad.TimeSignature):
                abjad.detach(abjad.TimeSignature, leaf)

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self) -> abjad.Container:
        r"""The |abjad.Container| to be hocketed."""
        return abjad.mutate.copy(self._contents)

    @contents.setter
    def contents(
        self,
        contents: abjad.Container,
    ) -> None:
        if not isinstance(contents, abjad.Container):
            raise TypeError("'contents' must be 'abjad.Container' or child class")
        if not abjad.select(contents).leaves().are_contiguous_logical_voice():
            raise ValueError("'contents' must be contiguous logical voice")
        if isinstance(contents, abjad.Score):
            self._contents = abjad.mutate.copy(contents[0])
        elif isinstance(contents, abjad.Tuplet):
            self._contents = abjad.Container([abjad.mutate.copy(contents)])
        else:
            self._contents = abjad.mutate.copy(contents)
        self._time_signatures = get.time_signature_list(
            contents,
            do_not_use_none=True,
        )

    @property
    def n_voices(self) -> int:
        r"""Number of individual voices in the output."""
        return self._n_voices

    @n_voices.setter
    def n_voices(
        self,
        n_voices: int,
    ) -> None:
        if not isinstance(n_voices, int):
            raise TypeError("'n_voices' must be 'int'")
        if n_voices < 1:
            raise ValueError("'n_voices' must be greater than zero")
        if self._force_k_voices and self._k > n_voices:
            raise ValueError(
                "'n_voices' cannot be smaller than 'k' when 'force_k_voices' is set to True"
            )
        self._n_voices = n_voices
        self._pitch_ranges = None

    @property
    def weights(self) -> list[Union[float, int]]:
        r"""The :obj:`list` with weights for each voice."""
        return self._weights

    @weights.setter
    def weights(
        self,
        weights: Optional[list[Union[float, int]]],
    ) -> None:
        if weights is not None:
            if not isinstance(weights, list):
                raise TypeError("'weights' must be 'list'")
            if not self.__len__() == len(weights):
                raise ValueError("'weights' must have the same length as 'n_voices'")
            if not all(isinstance(weight, (int, float)) for weight in weights):
                raise TypeError("'weights' elements must be 'int' or 'float'")
            self._weights = weights[:]
        else:
            self.reset_weights()

    @property
    def k(self) -> int:
        r"""Number of random choice operations applied to a logical tie."""
        return self._k

    @k.setter
    def k(
        self,
        k: int,
    ) -> None:
        if not isinstance(k, int):
            raise TypeError("'k' must be 'int'")
        if k < 1:
            raise ValueError("'k' must be greater than zero")
        if self._force_k_voices and k > self._n_voices:
            raise ValueError(
                "'k' cannot be greater than 'n_voices' when 'force_k_voices' is set to True"
            )
        self._k = k

    @property
    def pitch_ranges(self) -> list:
        r"""List of tuples or lists for the pitch ranges of each voice.
        Use the format:

        [(min0, max0), (min1, max1), (min2, max2), ...]

        ... where the indices are the voice numbers.
        """
        return self._pitch_ranges

    @pitch_ranges.setter
    def pitch_ranges(
        self,
        pitch_ranges: Optional[list],
    ) -> None:
        if pitch_ranges is not None:
            if not isinstance(pitch_ranges, list):
                raise TypeError("'pitch_ranges' must be 'list'")
            if len(pitch_ranges) != self._n_voices:
                raise ValueError("'pitch_ranges' must have length 'n_voices'")
            for voice_range in pitch_ranges:
                if not isinstance(voice_range, abjad.PitchRange):
                    raise TypeError("elements of 'pitch_ranges' must be 'abjad.PitchRange'")
        self._pitch_ranges = pitch_ranges

    @property
    def explode_chords(self) -> bool:
        r"""When ``True``, the hocket process will consider each note of a
        chord individually, 'exploding' it into several voices.
        """
        return self._explode_chords

    @explode_chords.setter
    def explode_chords(
        self,
        explode_chords: bool,
    ) -> None:
        if not isinstance(explode_chords, bool):
            raise TypeError("'explode_chords' must be 'bool'")
        self._explode_chords = explode_chords

    @property
    def force_k_voices(self) -> bool:
        r"""When ``True``, the hocket process will ensure that each logical tie
        is distributed among :attr:`k` voices.
        """
        return self._force_k_voices

    @force_k_voices.setter
    def force_k_voices(
        self,
        force_k_voices: bool,
    ) -> None:
        if not isinstance(force_k_voices, bool):
            raise TypeError("'force_k_voices' must be 'bool'")
        if force_k_voices and self._k > self._n_voices:
            raise ValueError(
                "'force_k_voices' cannot be set to True if 'k' > 'n_voices', change 'k' first"
            )
        self._force_k_voices = force_k_voices

    @property
    def omit_time_signatures(self) -> bool:
        r"""When ``True``, all time signatures will be omitted from the
        output.
        """
        return self._omit_time_signatures

    @omit_time_signatures.setter
    def omit_time_signatures(
        self,
        omit_time_signatures: bool,
    ) -> None:
        if not isinstance(omit_time_signatures, bool):
            raise TypeError("'omit_time_signatures' must be 'bool'")
        self._omit_time_signatures = omit_time_signatures

    @property
    def disable_rewrite_meter(self) -> bool:
        r"""When ``True``, the durations of the notes in the output will not be
        rewritten by the |abjad.Meter.rewrite_meter()| mutation. Rests will
        have the same duration as the logical ties they replaced.
        """
        return self._disable_rewrite_meter

    @disable_rewrite_meter.setter
    def disable_rewrite_meter(
        self,
        disable_rewrite_meter: bool,
    ) -> None:
        if not isinstance(disable_rewrite_meter, bool):
            raise TypeError("'disable_rewrite_meter' must be 'bool'")
        self._disable_rewrite_meter = disable_rewrite_meter

    @property
    def use_multimeasure_rests(self) -> bool:
        r"""When ``True``, multi-measure rests will be used for silent
        measures.
        """
        return self._use_multimeasure_rests

    @use_multimeasure_rests.setter
    def use_multimeasure_rests(
        self,
        use_multimeasure_rests: bool,
    ) -> None:
        if not isinstance(use_multimeasure_rests, bool):
            raise TypeError("'use_multimeasure_rests' must be 'bool'")
        self._use_multimeasure_rests = use_multimeasure_rests

    @property
    def boundary_depth(self) -> Union[int, None]:
        r"""Sets the argument ``boundary_depth`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._boundary_depth

    @boundary_depth.setter
    def boundary_depth(
        self,
        boundary_depth: Optional[int],
    ) -> None:
        if boundary_depth is not None:
            if not isinstance(boundary_depth, int):
                raise TypeError("'boundary_depth' must be 'int'")
        self._boundary_depth = boundary_depth

    @property
    def maximum_dot_count(self) -> Union[int, None]:
        r"""Sets the argument ``maximum_dot_count`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._maximum_dot_count

    @maximum_dot_count.setter
    def maximum_dot_count(
        self,
        maximum_dot_count: Optional[int],
    ) -> None:
        if maximum_dot_count is not None:
            if not isinstance(maximum_dot_count, int):
                raise TypeError("'maximum_dot_count' must be 'int'")
        self._maximum_dot_count = maximum_dot_count

    @property
    def rewrite_tuplets(self) -> bool:
        r"""Sets the argument ``rewrite_tuplets`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._rewrite_tuplets

    @rewrite_tuplets.setter
    def rewrite_tuplets(
        self,
        rewrite_tuplets: bool,
    ) -> None:
        if not isinstance(rewrite_tuplets, bool):
            raise TypeError("'rewrite_tuplets' must be 'bool'")
        self._rewrite_tuplets = rewrite_tuplets

    @property
    def prettify_rewrite_meter(self) -> bool:
        r"""Used to enable or disable the mutation
        |auxjad.mutate.prettify_rewrite_meter()| (default ``True``).
        """
        return self._prettify_rewrite_meter

    @prettify_rewrite_meter.setter
    def prettify_rewrite_meter(
        self,
        prettify_rewrite_meter: bool,
    ) -> None:
        if not isinstance(prettify_rewrite_meter, bool):
            raise TypeError("'prettify_rewrite_meter' must be 'bool'")
        self._prettify_rewrite_meter = prettify_rewrite_meter

    @property
    def extract_trivial_tuplets(self) -> bool:
        r"""Sets the argument ``extract_trivial_tuplets`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._extract_trivial_tuplets

    @extract_trivial_tuplets.setter
    def extract_trivial_tuplets(
        self,
        extract_trivial_tuplets: bool,
    ) -> None:
        if not isinstance(extract_trivial_tuplets, bool):
            raise TypeError("'extract_trivial_tuplets' must be 'bool'")
        self._extract_trivial_tuplets = extract_trivial_tuplets

    @property
    def fuse_across_groups_of_beats(self) -> bool:
        r"""Sets the argument ``fuse_across_groups_of_beats`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_across_groups_of_beats

    @fuse_across_groups_of_beats.setter
    def fuse_across_groups_of_beats(
        self,
        fuse_across_groups_of_beats: bool,
    ) -> None:
        if not isinstance(fuse_across_groups_of_beats, bool):
            raise TypeError("'fuse_across_groups_of_beats' must be 'bool'")
        self._fuse_across_groups_of_beats = fuse_across_groups_of_beats

    @property
    def fuse_quadruple_meter(self) -> bool:
        r"""Sets the argument ``fuse_quadruple_meter`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_quadruple_meter

    @fuse_quadruple_meter.setter
    def fuse_quadruple_meter(
        self,
        fuse_quadruple_meter: bool,
    ) -> None:
        if not isinstance(fuse_quadruple_meter, bool):
            raise TypeError("'fuse_quadruple_meter' must be 'bool'")
        self._fuse_quadruple_meter = fuse_quadruple_meter

    @property
    def fuse_triple_meter(self) -> bool:
        r"""Sets the argument ``fuse_triple_meter`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_triple_meter

    @fuse_triple_meter.setter
    def fuse_triple_meter(
        self,
        fuse_triple_meter: bool,
    ) -> None:
        if not isinstance(fuse_triple_meter, bool):
            raise TypeError("'fuse_triple_meter' must be 'bool'")
        self._fuse_triple_meter = fuse_triple_meter

    @property
    def current_window(self) -> Union[tuple, None]:
        r"""Read-only property, returns the result of the last operation as a
        :obj:`tuple` of |abjad.Selection|..
        """
        if self._voices is not None:
            output = []
            for voice in self._voices:
                voice_ = abjad.mutate.copy(voice)
                if self._omit_time_signatures:
                    self._remove_all_time_signatures(voice_)
                output.append(abjad.select([voice_]))
            return tuple(output[:])
        else:
            return self._voices



import random
from typing import Any, Union

import abjad

from .. import get, mutate


class _LooperParent:
    r"""This is the parent class of all Looper classes. It implements all
    common methods and properties, and initialises those using their @setter
    methods.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_contents",
        "_head_position",
        "_window_size",
        "_step_size",
        "_max_steps",
        "_repetition_chance",
        "_forward_bias",
        "_current_window",
        "_is_first_window",
        "_process_on_first_call",
    )

    ### INITIALISER ###

    def __init__(
        self,
        *,
        head_position: Any,
        window_size: Any,
        step_size: Any,
        max_steps: int = 1,
        repetition_chance: float = 0.0,
        forward_bias: float = 1.0,
        process_on_first_call: bool = False,
    ) -> None:
        if not isinstance(process_on_first_call, bool):
            raise TypeError("'process_on_first_call' must be 'bool'")
        self.head_position = head_position
        self.window_size = window_size
        self.step_size = step_size
        self.max_steps = max_steps
        self.repetition_chance = repetition_chance
        self.forward_bias = forward_bias
        self.process_on_first_call = process_on_first_call
        self._is_first_window = True
        self._current_window = None

    ### SPECIAL METHODS ###

    def __call__(self) -> abjad.Selection:
        r"""Calls the looping process for one iteration, returning an
        |abjad.Selection|.
        """
        self._move_head()
        if self._done:
            raise RuntimeError("'contents' has been exhausted")
        self._slice_contents()
        return self.current_window

    def __next__(self) -> abjad.Selection:
        r"""Calls the looping process for one iteration, returning an
        |abjad.Selection|.
        """
        self._move_head()
        if self._done:
            raise StopIteration
        self._slice_contents()
        return self.current_window

    def __iter__(self) -> None:
        r"""Returns an iterator, allowing instances to be used as iterators."""
        return self

    ### PUBLIC METHODS ###

    def output_all(
        self,
        *,
        tie_identical_pitches: bool = False,
    ) -> abjad.Selection:
        r"""Goes through the whole looping process and outputs a single
        |abjad.Selection|.
        """
        if not isinstance(tie_identical_pitches, bool):
            raise TypeError("'tie_identical_pitches' must be 'bool'")
        dummy_container = abjad.Container()
        while True:
            try:
                if not tie_identical_pitches or len(dummy_container) == 0:
                    dummy_container.append(self.__call__())
                else:
                    new_window = self.__call__()
                    leaf1 = abjad.select(new_window).leaf(0)
                    leaf2 = abjad.select(dummy_container).leaf(-1)
                    if get.leaves_are_tieable((leaf1, leaf2)):
                        abjad.attach(abjad.Tie(), dummy_container[-1])
                    dummy_container.append(new_window)
            except RuntimeError:
                break
        mutate.remove_repeated_time_signatures(dummy_container[:])
        mutate.reposition_dynamics(dummy_container[:])
        output = dummy_container[:]
        dummy_container[:] = []
        return output

    def output_n(
        self,
        n: int,
        *,
        tie_identical_pitches: bool = False,
    ) -> abjad.Selection:
        r"""Goes through ``n`` iterations of the looping process and outputs a
        single |abjad.Selection|.
        """
        if not isinstance(n, int):
            raise TypeError("first positional argument must be 'int'")
        if n < 1:
            raise ValueError("first positional argument must be a positive 'int'")
        if not isinstance(tie_identical_pitches, bool):
            raise TypeError("'tie_identical_pitches' must be 'bool'")
        dummy_container = abjad.Container()
        for _ in range(n):
            if not tie_identical_pitches or len(dummy_container) == 0:
                dummy_container.append(self.__call__())
            else:
                new_window = self.__call__()
                leaf1 = abjad.select(new_window).leaf(0)
                leaf2 = abjad.select(dummy_container).leaf(-1)
                if get.leaves_are_tieable((leaf1, leaf2)):
                    abjad.attach(abjad.Tie(), dummy_container[-1])
                dummy_container.append(new_window)
        mutate.remove_repeated_time_signatures(dummy_container[:])
        mutate.reposition_dynamics(dummy_container[:])
        output = dummy_container[:]
        dummy_container[:] = []
        return output

    ### PRIVATE METHODS ###

    def _move_head(self) -> None:
        r"""Moves the head by a certain number of steps of fixed size, either
        forwards or backwards according to the forward bias.
        """
        if not self._is_first_window or self._process_on_first_call:
            if self._repetition_chance == 0.0 or random.random() > self._repetition_chance:
                step = self._step_size * random.randint(1, self._max_steps)
                diretion = self._biased_choice(self._forward_bias)
                self._head_position += step * diretion
        self._is_first_window = False

    def _slice_contents(self) -> None:
        r"""Slices :attr:`contents`, will be defined for each individual child
        class.
        """
        pass

    @staticmethod
    def _biased_choice(
        forward_bias: float,
    ) -> None:
        r"""Returns either +1 or -1 according to a forward bias value."""
        weights = [forward_bias, 1.0 - forward_bias]
        return random.choices([1, -1], weights=weights)[0]

    @staticmethod
    def _remove_all_time_signatures(
        container: abjad.Container,
    ) -> None:
        r"""Removes all time signatures of an |abjad.Container|."""
        for leaf in abjad.select(container).leaves():
            if abjad.get.effective(leaf, abjad.TimeSignature):
                abjad.detach(abjad.TimeSignature, leaf)

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self) -> None:
        r""":attr:`contents` property will be defined for each individual child
        class.
        """
        pass

    @contents.setter
    def contents(self) -> None:
        pass

    @property
    def head_position(self) -> int:
        r"""The position of the head at the start of a looping window."""
        return self._head_position

    @head_position.setter
    def head_position(
        self,
        head_position: int,
    ) -> None:
        if not isinstance(head_position, int):
            raise TypeError("'head_position' must be 'int'")
        if head_position < 0:
            raise ValueError("'head_position' must be a positive 'int'")
        if head_position >= self.__len__():
            raise ValueError("'head_position' must be smaller than the length of 'contents'")
        self._is_first_window = True
        self._head_position = head_position

    @property
    def window_size(self) -> int:
        r"""The length of the looping window."""
        return self._window_size

    @window_size.setter
    def window_size(
        self,
        window_size: int,
    ) -> None:
        if not isinstance(window_size, int):
            raise TypeError("'window_size' must be 'int'")
        if window_size < 1:
            raise ValueError("'window_size' must be greater than zero")
        if window_size > self.__len__():
            raise ValueError(
                "'window_size' must be smaller than or equal to the length of 'contents'"
            )
        self._window_size = window_size

    @property
    def step_size(self) -> int:
        r"""The size of each step when moving the head."""
        return self._step_size

    @step_size.setter
    def step_size(
        self,
        step_size: int,
    ) -> None:
        if not isinstance(step_size, int):
            raise TypeError("'step_size' must be 'int'")
        if step_size < 1:
            raise ValueError("'step_size' must be greater than zero")
        if step_size >= self.__len__():
            raise ValueError("'step_size' must be smaller than the length of 'contents'")
        self._step_size = step_size

    @property
    def max_steps(self) -> int:
        r"""The maximum number of steps per operation."""
        return self._max_steps

    @max_steps.setter
    def max_steps(
        self,
        max_steps: int,
    ) -> None:
        if not isinstance(max_steps, int):
            raise TypeError("'max_steps' must be 'int'")
        if max_steps < 1:
            raise ValueError("'max_steps' must be greater than zero")
        self._max_steps = max_steps

    @property
    def repetition_chance(self) -> float:
        r"""The chance of the head not moving, thus repeating the output."""
        return self._repetition_chance

    @repetition_chance.setter
    def repetition_chance(
        self,
        repetition_chance: float,
    ) -> None:
        if not isinstance(repetition_chance, float):
            raise TypeError("'repetition_chance' must be 'float'")
        if repetition_chance < 0.0 or repetition_chance > 1.0:
            raise ValueError("'repetition_chance' must be between 0.0 and 1.0")
        self._repetition_chance = repetition_chance

    @property
    def forward_bias(self) -> float:
        r"""The chance of the window moving forward instead of backwards. It
        should range from ``0.0`` to ``1.0`` (default ``1.0``, which means the
        window can only move forwards. A value of ``0.5`` gives :math:`50\%`
        chance of moving forwards while a value of ``0.0`` will move the window
        only backwards).
        """
        return self._forward_bias

    @forward_bias.setter
    def forward_bias(
        self,
        forward_bias: float,
    ) -> None:
        if not isinstance(forward_bias, float):
            raise TypeError("'forward_bias' must be 'float'")
        if forward_bias < 0.0 or forward_bias > 1.0:
            raise ValueError("'forward_bias' must be between 0.0 and 1.0")
        self._forward_bias = forward_bias

    @property
    def process_on_first_call(self) -> bool:
        r"""If ``True`` then :attr:`contents` will be processed in the very
        first call.
        """
        return self._process_on_first_call

    @process_on_first_call.setter
    def process_on_first_call(
        self,
        process_on_first_call: bool,
    ) -> None:
        if not isinstance(process_on_first_call, bool):
            raise TypeError("'process_on_first_call' must be 'bool'")
        self._process_on_first_call = process_on_first_call

    @property
    def current_window(self) -> Union[abjad.Selection, None]:
        r"""Read-only property, returns the window at the current head
        position.
        """
        if self._current_window is None:
            return self._current_window
        current_window = abjad.mutate.copy(self._current_window)
        if self._omit_time_signatures:
            self._remove_all_time_signatures(current_window)
        return current_window

    ### PRIVATE PROPERTIES ###

    @property
    def _done(self) -> bool:
        r""":obj:`bool` indicating whether the process is done (i.e. whether
        the head position has overtaken the :attr:`contents`'s length).
        """
        return self._head_position >= self.__len__() or self._head_position < 0



import random
from typing import Any, Union


class CartographySelector:
    r"""A selector used to store, manipulate, and select objects using a
    weighted function constructed with a fixed decay rate. The decay rate
    represents the ratio of probabilities of any index given the probability of
    the preceding one. For instance, if the decay rate is set to ``0.75``
    (which is its default value), the probability of the element in index ``1``
    of the input :obj:`list` being selected is :math:`75\%`` the probability of
    the element in index ``0``, and the probability of the element in index
    ``2`` is :math:`56.25\%`` (i.e. :math:`0.75^2`) the probability of the
    element in index ``0``. The probability :math:`P(n)`` of the :math:`n`-th
    element can thus be expressed as a relation to the probability of another
    element :math:`k` indexes apart using:

    ..  math::

        P(n) = (3/4)^k \times P(n-k)

    This is the selector used in my *Cartography* series of compositions.

    Basic usage:
        The selector should be initialised with a :obj:`list` of objects. The
        elements of this :obj:`list` can be of any type.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.contents
        [0, 1, 2, 3, 4]

        Calling the selector will output one of its elements, selected
        according to its weight function.

        >>> selector()
        2

        The default decay rate is ``0.75``; that is, the weight of any given
        elements is the weight of the previous one multiplied by ``0.75``. The
        :attr:`weights` are associated with the index position, not the
        elements themselves.

        >>> selector.weights
        [1.0, 0.75, 0.5625, 0.421875, 0.31640625]

        By default, only the weight function (defined by the decay rate) is
        taken into consideration when selecting an element. This means that
        repeated elements can appear, as shown below.

        >>> result = ''
        >>> for _ in range(30):
        ...     result += str(selector())
        >>> result
        203001402200011111101400310140

    :func:`len()` function:
        Applying the :func:`len()` function to the selector will return the
        length of the input :obj:`list`.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> len(selector)
        5

    :func:`next()` function:
        Alternatively, use the :func:`next()` function or :meth:`__next__()`
        method to get the next result.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.__next__()
        1
        >>> next(selector)
        0

    :meth:`__call__` and argument ``no_repeat``:
        Calling the selector with the optional keyword argument ``no_repeat``
        set to ``True`` will forbid immediate repetitions among consecutive
        calls.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> result = ''
        >>> for _ in range(30):
        ...     result += str(selector(no_repeat=True))
        >>> result
        210421021020304024230120241202

    :attr:`decay_rate`:
        The keyword argument :attr:`decay_rate` can be used to set a different
        decay rate when creating a selector.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4],
        ...                                       decay_rate=0.5,
        ...                                       )
        >>> selector.weights
        [1.0, 0.5, 0.25, 0.125, 0.0625]

        The decay rate can also be set after the creation of a selector using,
        the property :attr:`decay_rate`.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.decay_rate = 0.2
        >>> selector.weights
        [1.0, 0.2, 0.04000000000000001, 0.008000000000000002,
        0.0016000000000000003]
        >>> result = ''
        >>> for _ in range(30):
        ...     result += str(selector())
        >>> result
        '000001002100000201001030000100'

    :meth:`drop_first_and_append`:
        This is a type of content transformation, it drops the first element of
        :attr:`contents`, shifts all others leftwards, and appends the new
        element to the last index.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.drop_first_and_append(5)
        >>> selector.contents
        [1, 2, 3, 4, 5]
        >>> selector.drop_first_and_append(42)
        >>> selector.contents
        [2, 3, 4, 5, 42]

    :meth:`drop_n_and_append`:
        This is a type of content transformation similar to
        :meth:`drop_first_and_append`, it drops the element at index ``n`` of
        :attr:`contents`, shifts all the next elements one position lefwards,
        and appends the new element at the last index.

        >>> selector = auxjad.CartographySelector([10, 7, 14, 31, 98])
        >>> selector.contents
        [10, 7, 14, 31, 98]
        >>> selector.drop_n_and_append(100, n=2)
        >>> selector.contents
        [10, 7, 31, 98, 100]

    :meth:`drop_last_and_prepend`:
        A type of content transformation, it drops the last element of
        :attr:`contents`, shifts all others rightwards, and then prepends
        the new element to the first index.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.drop_last_and_prepend(-1)
        >>> selector.contents
        [-1, 0, 1, 2, 3]
        >>> selector.drop_last_and_prepend(71)
        >>> selector.contents
        [71, -1, 0, 1, 2]

    :meth:`rotate`:
        Rotation is another type of content transformation. It rotates all
        elements rightwards, moving the last element to the first index. If the
        optional keyword argument ``anticlockwise`` is set to ``True``, the
        rotation will be in the opposite direction.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.rotate()
        >>> selector.contents
        [1, 2, 3, 4, 0]
        >>> selector.rotate(anticlockwise=True)
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.rotate(anticlockwise=True)
        >>> selector.contents
        [1, 2, 3, 4, 0]

    :meth:`mirror_swap`:
        The mirror swap transformation swaps takes an input index and swaps the
        element at tit with its complementary element. Complementary elements
        are defined as the pair of elements which share the same distance from
        the centre of the :attr:`contents` (in terms of number of indeces), and
        are located at either side of this centre.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.mirror_swap(0)
        >>> selector.contents
        [4, 1, 2, 3, 0]
        >>> selector.mirror_swap(0)
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.mirror_swap(3)
        >>> selector.contents
        [0, 3, 2, 1, 4]
        >>> selector.mirror_swap(2)
        >>> selector.contents
        [0, 3, 2, 1, 4]

    :meth:`mirror_random_swap`:
        A type of content transformation which will apply  :meth:`mirror_swap`
        to a random pair of complementary elements. In case of a selector with
        an odd number of elements, this method will never pick the element at
        the central index since that is the pivot point of the operation and it
        would not result in any changes.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.mirror_random_swap()
        >>> selector.contents
        [4, 1, 2, 3, 0]
        >>> selector.mirror_random_swap()
        >>> selector.contents
        [4, 3, 2, 1, 0]
        >>> selector.mirror_random_swap()
        >>> selector.contents
        [4, 1, 2, 3, 0]

    :meth:`shuffle`:
        The method :meth:`shuffle` will shuffle the position of the elements of
        the selector's :attr:`contents`.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.shuffle()
        >>> selector.contents
        [1, 4, 3, 0, 2]

    :attr:`contents`:
        The contents of a selector can also be altered after it has been
        initialised using the :attr:`contents` property. The length of the
        contents can change as well.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4],
        ...                                       decay_rate=0.5,
        ...                                       )
        >>> len(selector)
        5
        >>> selector.weights
        [1.0, 0.5, 0.25, 0.125, 0.0625]
        >>> selector.contents = [10, 7, 14, 31, 98, 47, 32]
        >>> selector.contents
        [10, 7, 14, 31, 98, 47, 32]
        >>> len(selector)
        7
        >>> selector.weights
        [1.0, 0.5, 0.25, 0.125, 0.0625, 0.03125, 0.015625]

    :attr:`previous_result` and :attr:`previous_index`:
        Use the read-only properties :attr:`previous_result` and
        :attr:`previous_index` to output the previous result and its index.

        >>> selector = auxjad.CartographySelector([10, 7, 14, 31, 98])
        >>> selector()
        14
        >>> previous_index = selector.previous_index
        >>> previous_index
        2
        >>> selector.previous_result
        14

    Slicing and indexing:
        Instances of this class can be indexed and sliced. This allows reading,
        assigning, or deleting values from :attr:`contents`.

        >>> selector = auxjad.CartographySelector([10, 7, 14, 31, 98])
        >>> selector[1]
        7
        >>> selector[1:4]
        [7, 14, 31]
        >>> selector[:]
        [10, 7, 14, 31, 98]
        >>> selector()
        31
        >>> previous_index = selector.previous_index
        >>> previous_index
        3
        >>> selector[previous_index]
        31
        >>> selector.contents
        [10, 7, 14, 31, 98]
        >>> selector[2] = 100
        >>> selector.contents
        [10, 7, 100, 31, 98]
        >>> del selector[2:4]
        >>> selector.contents
        [10, 7, 98]
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_contents",
        "_decay_rate",
        "_previous_index",
        "_weights",
    )

    ### INITIALISER ###

    def __init__(
        self,
        contents: list[Any],
        *,
        decay_rate: float = 0.75,
    ) -> None:
        if not isinstance(contents, list):
            raise TypeError("'contents' must be 'list'")
        if not isinstance(decay_rate, float):
            raise TypeError("'decay_rate' must be 'float'")
        if decay_rate <= 0.0 or decay_rate > 1.0:
            raise ValueError("'decay_rate' must be larger than 0.0 and less than or equal to 1.0")
        self._contents = contents[:]
        self._previous_index = None
        self._decay_rate = decay_rate
        self._generate_weights()

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r"""Returns interpreter representation of :attr:`contents`."""
        return str(self._contents)

    def __len__(self) -> int:
        r"""Returns the length of :attr:`contents`."""
        return len(self._contents)

    def __call__(
        self,
        *,
        no_repeat: bool = False,
    ) -> Any:
        r"""Calls the selection process and outputs one element of
        :attr:`contents`.
        """
        if not isinstance(no_repeat, bool):
            raise TypeError("'no_repeat' must be 'bool")
        if not no_repeat:
            new_index = random.choices(
                [n for n in range(self.__len__())],
                weights=self._weights,
            )[0]
        else:
            new_index = self._previous_index
            while new_index == self._previous_index:
                new_index = random.choices(
                    [n for n in range(self.__len__())],
                    weights=self._weights,
                )[0]
        self._previous_index = new_index
        return self._contents[self._previous_index]

    def __next__(self) -> Any:
        r"""Calls the selection process and outputs one element of
        :attr:`contents`.
        """
        return self.__call__()

    def __getitem__(
        self,
        key: int,
    ) -> Any:
        r"""Returns one or more elements of :attr:`contents` through indexing
        or slicing.
        """
        return self._contents[key]

    def __setitem__(
        self,
        key: int,
        value: Any,
    ) -> None:
        r"""Assigns values to one or more elements of :attr:`contents` through
        indexing or slicing.
        """
        self._contents[key] = value
        self._generate_weights()

    def __delitem__(
        self,
        key: int,
    ) -> None:
        r"""Deletes one or more elements of :attr:`contents` through indexing
        or slicing.
        """
        del self._contents[key]
        self._generate_weights()

    ### PUBLIC METHODS ###

    def drop_first_and_append(
        self,
        new_element: Any,
    ) -> None:
        r"""A type of content transformation, it drops the first element of
        :attr:`contents`, shifts all others leftwards, and appends the new
        element to the last index.
        """
        self._contents = self._contents[1:] + [new_element]

    def drop_n_and_append(
        self,
        new_element: Any,
        *,
        n: int,
    ) -> None:
        r"""A type of content transformation similar to
        :meth:`drop_first_and_append`, it drops the element at index ``n`` of
        :attr:`contents`, shifts all the next elements one position lefwards,
        and appends the new element at the last index.
        """
        self._contents = self._contents[:n] + self._contents[n + 1 :] + [new_element]

    def drop_last_and_prepend(
        self,
        new_element: Any,
    ) -> None:
        r"""A type of content transformation, it drops the last element of
        :attr:`contents`, shifts all others rightwards, and then prepends
        the new element to the first index.
        """
        self._contents = [new_element] + self._contents[:-1]

    def rotate(
        self,
        *,
        anticlockwise=False,
    ) -> None:
        r"""A type of content transformation, it rotates all elements
        rightwards, moving the last element to the first index. If the optional
        keyword argument ``anticlockwise`` is set to ``True``, the rotation
        will be in the opposite direction.
        """
        if not anticlockwise:
            self._contents = self._contents[1:] + self._contents[:1]
        else:
            self._contents = self._contents[-1:] + self._contents[:-1]

    def mirror_swap(
        self,
        index: int,
    ) -> None:
        r"""A type of content transformation which takes an input index and
        swaps the element at that position with its complementary element.
        Complementary elements are defined as the pair of elements which share
        the same distance from the centre of the :attr:`contents` (in terms of
        number of indeces), and are located at either side of this centre.
        """
        self._contents[index], self._contents[-1 - index] = (
            self._contents[-1 - index],
            self._contents[index],
        )

    def mirror_random_swap(self) -> None:
        r"""A type of content transformation which will apply
        :meth:`mirror_swap` to a random pair of complementary elements. In case
        of a selector with an odd number of elements, this method will never
        pick the element at the central index since that is the pivot point of
        the operation and it would not result in any changes.
        """
        max_index = self.__len__() // 2 - 1
        self.mirror_swap(random.randint(0, max_index))

    def shuffle(self) -> None:
        r"""Shuffles the position of the elements of :attr:`contents`."""
        random.shuffle(self._contents)

    ### PRIVATE METHODS ###

    def _generate_weights(self) -> None:
        r"""Given a decay rate, this method generates the :attr:`weights` of
        individual indeces.
        """
        self._weights = []
        for n in range(self.__len__()):
            self._weights.append(self._decay_rate**n)

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self) -> list[Any]:
        r"""The :obj:`list` from which the selector picks elements."""
        return self._contents

    @contents.setter
    def contents(
        self,
        contents: list[Any],
    ) -> None:
        if not isinstance(contents, list):
            raise TypeError("'contents' must be 'list")
        self._contents = contents[:]
        self._generate_weights()

    @property
    def decay_rate(self) -> float:
        r"""The decay rate represents the ratio of probabilities of any index
        given the probability of the preceding one. For instance, if the decay
        rate is set to ``0.75`` (which is its default value), the probability
        of the element in index ``1`` of the input :obj:`list` being selected
        is ``0.75`` the probability of the element in index ``0``, and the
        probability of the element in index ``2`` is ``0.5625`` (i.e. ``0.75``
        squared) the probability of the element in index ``0``.
        """
        return self._decay_rate

    @decay_rate.setter
    def decay_rate(
        self,
        decay_rate: float,
    ) -> None:
        if not isinstance(decay_rate, float):
            raise TypeError("'decay_rate' must be float")
        if decay_rate <= 0.0 or decay_rate > 1.0:
            raise ValueError("'decay_rate' must be larger than 0.0 and less than or equal to 1.0")
        self._decay_rate = decay_rate
        self._generate_weights()

    @property
    def previous_index(self) -> Union[int, None]:
        r"""Read-only property, returns the index of the previously output
        element.
        """
        return self._previous_index

    @property
    def previous_result(self) -> Any:
        r"""Read-only property, returns the previously output element."""
        if self._previous_index is not None:
            return self._contents[self._previous_index]
        else:
            return self._previous_index

    @property
    def weights(self) -> list[float]:
        r"""Read-only property, returns the weight vector."""
        return self._weights



import abjad


def time_signature_list(
    container: abjad.Container,
    *,
    do_not_use_none: bool = False,
    implicit_common_time: bool = True,
    omit_repeated: bool = False,
) -> list[abjad.TimeSignature]:
    r"""Returns a :obj:`list` with the |abjad.TimeSignature|'s for all measures
    of an input |abjad.Container|.

    Basic usage:
        This function returns a :obj:`list` with one |abjad.TimeSignature| per
        measure.

        >>> container = abjad.Container(r"\time 3/4 c'2. \time 4/4 e'1")
        >>> time_signatures = auxjad.get.time_signature_list(container)
        >>> time_signatures
        [TimeSignature((3, 4)), TimeSignature((4, 4))]

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.get|. It can thus be used from either |auxjad.get|_ or
        |abjad.get| namespaces. Therefore, the two lines below are equivalent:

        >>> container = abjad.Container(r"\time 3/4 c'2. \time 4/4 e'1")
        >>> auxjad.get.time_signature_list(container)
        [TimeSignature((3, 4)), TimeSignature((4, 4))]
        >>> abjad.get.time_signature_list(container)
        [TimeSignature((3, 4)), TimeSignature((4, 4))]

    ``do_not_use_none``:
        By default, the :obj:`list` will contain a ``None`` if a measure does
        not have an explicit time signature.

        >>> container = abjad.Container(
        ...     r"\time 5/8 c'4 ~ c'16 \time 3/8 d'4. e'4."
        ... )
        >>> time_signatures = auxjad.get.time_signature_list(container)
        >>> time_signatures
        [TimeSignature((5, 8)), TimeSignature((3, 8)), None]

        Set the keyword argument ``do_not_use_none`` to ``True`` to change this
        behaviour.

        >>> time_signatures = auxjad.get.time_signature_list(
        ...     container,
        ...     do_not_use_none=True,
        ... )
        >>> time_signatures
        [TimeSignature((5, 8)), TimeSignature((3, 8)), TimeSignature((3, 8))]

    ``omit_repeated``:
        By default, time signatures are output according to the container, even
        if there are multiple instances of a same time signature.

        >>> container = abjad.Container(
        ...     r"\time 3/4 c'2. d'2. \time 3/4 e'2. f'2."
        ... )
        >>> time_signatures = auxjad.get.time_signature_list(container)
        >>> time_signatures
        [abjad.TimeSignature((3, 4)), None, abjad.TimeSignature((3, 4)), None]

        Set the keyword argument ``omit_repeated`` to ``True`` to replace
        repeated time signatures with ``None``.

        >>> time_signatures = auxjad.get.time_signature_list(
        ...     container,
        ...     omit_repeated=True,
        ... )
        >>> time_signatures
        [abjad.TimeSignature((3, 4)), None, None, None]

    ..  error::

        Setting both ``do_not_use_none`` and ``omit_repeated`` to ``True`` will
        raise a :exc:`ValueError` exception:

        >>> container = abjad.Container(
        ...     r"\time 3/4 c'2. d'2. \time 3/4 e'2. f'2."
        ... )
        >>> time_signatures = auxjad.get.time_signature_list(
        ...     container,
        ...     do_not_use_none=True,
        ...     omit_repeated=True,
        ... )
        ValueError: 'omit_repeated' and 'do_not_use_none' cannot be both set to
        'True'

    ``implicit_common_time``:
        LilyPond uses an implicit time signature of 4/4 whenever a time
        signature is not found. This function behaves the same way.

        >>> container = abjad.Container(r"c'1 d'1 e'1 f'1")
        >>> time_signatures = auxjad.get.time_signature_list(container)
        >>> time_signatures
        [TimeSignature((4, 4)), None, None, None]

        To disable this behaviour, set ``implicit_common_time`` to ``False``.

        >>> time_signatures = auxjad.get.time_signature_list(
        ...     container,
        ...     implicit_common_time=False,
        ... )
        >>> time_signatures
        [None, None, None, None]

    ..  error::

        Setting ``do_not_use_none`` to ``True`` and ``implicit_common_time`` to
        ``False`` on a container that starts with no time signature will raise
        a :exc:`ValueError` exception:

        >>> container = abjad.Container(r"c'1 d'1 e'1 f'1")
        >>> time_signatures = auxjad.get.time_signature_list(
        ...     container,
        ...     do_not_use_none=True,
        ...     implicit_common_time=False,
        ... )
        ValueError: container does not have a time signature attached to its
        first leaf, with 'implicit_common_time' set to 'False' and
        'omit_repeated' set to 'True'
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("first argument must be 'abjad.Container' or child class")
    if not isinstance(do_not_use_none, bool):
        raise TypeError("'do_not_use_none' must be 'bool'")
    if not isinstance(implicit_common_time, bool):
        raise TypeError("'implicit_common_time' must be 'bool'")
    if not isinstance(omit_repeated, bool):
        raise TypeError("'omit_repeated' must be 'bool'")
    if omit_repeated and do_not_use_none:
        raise ValueError("'omit_repeated' and 'do_not_use_none' cannot be both set to 'True'")
    if not implicit_common_time and do_not_use_none:
        head_leaf = abjad.select(container).leaf(0)
        if not abjad.get.indicator(head_leaf, abjad.TimeSignature):
            raise ValueError(
                "container does not have a time signature "
                "attached to its first leaf, with "
                "'implicit_common_time' set to 'False' and "
                "'omit_repeated' set to 'True'"
            )

    measures = abjad.select(container[:]).group_by_measure()
    time_signatures = []

    for measure in measures:
        head = abjad.select(measure).leaf(0)
        time_signature = abjad.get.indicator(head, abjad.TimeSignature)
        if measure is measures[0] and time_signature is None:
            if implicit_common_time:
                time_signature = abjad.TimeSignature((4, 4))
        time_signatures.append(time_signature)

    if do_not_use_none:
        if time_signatures[0] is None:
            time_signatures[0] = abjad.TimeSignature((4, 4))
        for i in range(1, len(time_signatures)):
            if time_signatures[i] is None:
                time_signatures[i] = abjad.TimeSignature(time_signatures[i - 1].pair)
    elif omit_repeated:
        effective_time_signature = None
        for i in range(len(time_signatures)):
            current_time_signature = time_signatures[i]
            if current_time_signature is not None:
                current_time_signature = abjad.TimeSignature(current_time_signature.pair)
            if current_time_signature == effective_time_signature:
                time_signatures[i] = None
            else:
                if current_time_signature is not None:
                    effective_time_signature = current_time_signature

    return time_signatures



import abjad


def underfull_duration(selection: abjad.Selection) -> abjad.Duration:
    r"""Returns an |abjad.Duration| representing the duration missing in the
    last measure of an input |abjad.Selection| which is not fully filled in.

    Basic usage:
        Returns the missing duration of the last measure of an
        |abjad.Selection|. If no time signature is encountered, it uses
        LilyPond's fallback time signature of ``4/4``.

        >>> container1 = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Container(r"c'4 d'4 e'4")
        >>> container3 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4")
        >>> container4 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4 d'4 e'4 f'4")
        >>> auxjad.get.underfull_duration(container1[:])
        0
        >>> auxjad.get.underfull_duration(container2[:])
        1/4
        >>> auxjad.get.underfull_duration(container3[:])
        3/4
        >>> auxjad.get.underfull_duration(container4[:])
        0

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.get|. It can thus be used from either |auxjad.get|_ or
        |abjad.get| namespaces. Therefore, the two lines below are equivalent:

        >>> container = abjad.Container(r"c'4 d'4 e'4")
        >>> auxjad.get.underfull_duration(container[:])
        1/4
        >>> abjad.get.underfull_duration(container[:])
        1/4

    Time signature changes:
        Handles any time signatures as well as changes of time signature.

        >>> container1 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Container(r"\time 3/4 a2. \time 2/4 r2")
        >>> container3 = abjad.Container(r"\time 5/4 g1 ~ g4 \time 4/4 af'2")
        >>> container4 = abjad.Container(r"\time 6/8 c'2 ~ c'8")
        >>> auxjad.get.underfull_duration(container1[:])
        0
        >>> auxjad.get.underfull_duration(container2[:])
        0
        >>> auxjad.get.underfull_duration(container3[:])
        1/2
        >>> auxjad.get.underfull_duration(container4[:])
        1/8

    Partial time signatures:
        Correctly handles partial time signatures.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> time_signature = abjad.TimeSignature((3, 4), partial=(1, 4))
        >>> abjad.attach(time_signature, container[0])
        >>> auxjad.get.underfull_duration(container[:])
        0

    Multi-measure rests:
        It also handles multi-measure rests.

        >>> container1 = abjad.Container(r"R1")
        >>> container2 = abjad.Container(r"\time 3/4 R1 * 3/4 \time 2/4 r2")
        >>> container3 = abjad.Container(r"\time 5/4 R1 * 5/4 \time 4/4 g''4")
        >>> container4 = abjad.Container(r"\time 6/8 R1 * 1/2")
        >>> auxjad.get.underfull_duration(container1[:])
        0
        >>> auxjad.get.underfull_duration(container2[:])
        0
        >>> auxjad.get.underfull_duration(container3[:])
        3/4
        >>> auxjad.get.underfull_duration(container4[:])
        1/4

    ..  error::

        If a selection is malformed, i.e. it has an underfilled measure before
        a time signature change, the function raises a :exc:`ValueError`
        exception. This is also the case when a selection starts in the middle
        of a measure.

        >>> container = abjad.Container(r"\time 5/4 g''1 \time 4/4 f'1")
        >>> auxjad.get.underfull_duration(container[:])
        ValueError: 'selection' is malformed, with an underfull measure
        preceding a time signature change

    ..  warning::

        The input container must be a contiguous logical voice. When dealing
        with a container with multiple subcontainers (e.g. a score containing
        multiple staves), the best approach is to cycle through these
        subcontainers, applying this function to them individually.
    """
    if not isinstance(selection, abjad.Selection):
        raise TypeError("argument must be 'abjad.Selection'")
    if not selection.leaves().are_contiguous_logical_voice():
        raise ValueError("argument must be contiguous logical voice")
    leaves = selection.leaves()
    # handling first leaf
    time_signature = abjad.get.effective(leaves[0], abjad.TimeSignature)
    if time_signature is not None:
        effective_time_signature = time_signature
    else:
        effective_time_signature = abjad.TimeSignature((4, 4))
    duration = abjad.get.duration(leaves[0])
    # handling partial time signatures
    if effective_time_signature.partial is not None:
        duration += effective_time_signature.duration
        duration -= effective_time_signature.partial
    # all other leaves
    for leaf in leaves[1:]:
        time_signature = abjad.get.effective(leaf, abjad.TimeSignature)
        if time_signature is not None and time_signature != effective_time_signature:
            if duration % effective_time_signature.duration != 0:
                raise ValueError(
                    "'selection' is malformed, with an underfull "
                    "measure preceding a time signature change"
                )
            effective_time_signature = time_signature
            duration = abjad.Duration(0)
        duration += abjad.get.duration(leaf)
    duration_last_bar = duration % effective_time_signature.duration
    duration_left = duration_last_bar
    if duration_last_bar > abjad.Duration(0):
        duration_left = effective_time_signature.duration - duration_last_bar
    return duration_left



import abjad


def extract_trivial_tuplets(selection: abjad.Selection) -> None:
    r"""Mutates an input |abjad.Selection| in place and has no return value;
    this function looks for tuplets filled with rests or with tied notes or
    chords and replaces them with a single leaf.

    Basic usage:
        Usage is simple:

        >>> staff = abjad.Staff(
        ...     r"\times 2/3 {r4 r2} \times 2/3 {c'8 ~ c'8 ~ c'2}"
        ... )
        >>> abjad.show(container)

        ..  docs::

            {
                \times 2/3
                {
                    r4
                    r2
                }
                \times 2/3
                {
                    c'8
                    ~
                    c'8
                    ~
                    c'2
                }
            }

        ..  figure:: ../_images/extract_trivial_tuplets-4htz2xebxwf.png

        >>> auxjad.mutate(container[:]).extract_trivial_tuplets()
        >>> abjad.show(container)

        ..  docs::

            {
                r2
                c'2
            }

        ..  figure:: ../_images/extract_trivial_tuplets-2dbuwo4erhb.png

        It also works with containers with tuplets within tuplets.

        >>> container = abjad.Container(r"\times 4/5 {r2. \times 2/3 {r2 r4}}")
        >>> abjad.show(container)

        ..  docs::

            {
                \times 4/5
                {
                    r2.
                    \times 2/3
                    {
                        r2
                        r4
                    }
                }
            }

        ..  figure:: ../_images/extract_trivial_tuplets-8d5bcyxcmhc.png

        >>> auxjad.mutate(container[:]).extract_trivial_tuplets()
        >>> abjad.show(container)

        ..  docs::

            {
                r1
            }

        ..  figure:: ../_images/extract_trivial_tuplets-2a2fvwimyrx.png

        >>> container = abjad.Container(
        ...     r"\times 4/5 {c'2. ~ \times 2/3 {c'2 ~ c'4}}"
        ... )
        >>> abjad.show(container)

        ..  docs::

            {
                \times 4/5
                {
                    c'2.
                    ~
                    \times 2/3
                    {
                        c'2
                        ~
                        c'4
                    }
                }
            }

        ..  figure:: ../_images/extract_trivial_tuplets-xka6r5iyo4l.png

        >>> auxjad.mutate(staff[:]).extract_trivial_tuplets()
        >>> abjad.show(container)

        ..  docs::

            {
                c'1
            }

        ..  figure:: ../_images/extract_trivial_tuplets-f1qxi44xcsw.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate(staff[:]).extract_trivial_tuplets()
        >>> abjad.mutate(staff[:]).extract_trivial_tuplets()

    Partial extraction:
        This function also extracts tuplets within tuplets.

        >>> container = abjad.Container(
        ...     r"r2 \times 2/3 {r2 r4} \times 4/5 {c'2. \times 2/3 {r2 r4}}"
        ... )
        >>> abjad.show(container)

        ..  docs::

            {
                r2
                \times 2/3
                {
                    r2
                    r4
                }
                \times 4/5
                {
                    c'2.
                    \times 2/3
                    {
                        r2
                        r4
                    }
                }
            }

        ..  figure:: ../_images/extract_trivial_tuplets-adibnkb1mbs.png

        >>> auxjad.mutate(container[:]).extract_trivial_tuplets()
        >>> abjad.show(container)

        ..  docs::

            {
                r2
                r2
                \times 4/5
                {
                    c'2.
                    r2
                }
            }

        ..  figure:: ../_images/extract_trivial_tuplets-xldohyedqs.png

    ..  tip::

        Use |auxjad.mutate.rests_to_multimeasure_rest()| to replace measures
        filled with rests by a single multi-measure rest. That function makes
        use of |auxjad.mutate.extract_trivial_tuplets()|, so it is not
        necessary to flatten the empty tuplets beforehand.

    Time signature changes:
        Works with measures with any time signature.

        >>> container = abjad.Staff(r"\time 3/4 r2. \times 3/2 {r4 r4}")
        >>> auxjad.mutate(container[:]).extract_trivial_tuplets()
        >>> abjad.show(container)

        ..  docs::

            \new Staff
            {
                \time 3/4
                r2.
                r2.
            }

        ..  figure:: ../_images/extract_trivial_tuplets-sa1tqmvtkx.png

    Non-assignable durations:
        This function also extracts tuplets which sum up to a non-assignable
        duration. In this case, it creates multiple leaves and substitutes them
        for the original tuplet. Indicators are passed on to the first leaf of
        the new leaves.

        >>> staff = abjad.Staff(r"\time 6/4 c'4\f \times 5/6 {g1.\p}")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 6/4
                c'4
                \f
                \tweak text #tuplet-number::calc-fraction-text
                \times 5/6
                {
                    g1.
                    \p
                }
            }

        ..  figure:: ../_images/extract_trivial_tuplets-l4kp9g5v7m.png

        >>> abjad.mutate(staff[:]).extract_trivial_tuplets()
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 6/4
                c'4
                \f
                g1
                \p
                ~
                g4
            }

        ..  figure:: ../_images/extract_trivial_tuplets-8r40ndemvpn.png

    ..  note::

        When using |abjad.Container|'s, all time signatures in the output will
        be commented out with ``%%%.`` This is because Abjad only applies time
        signatures to containers that belong to a |abjad.Staff|. The present
        function works with either |abjad.Container| and |abjad.Staff|.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4")
        >>> abjad.show(container)

        ..  docs::

            {
                %%% \time 3/4 %%%
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/extract_trivial_tuplets-6wymsb7z1n4.png

        >>> staff = abjad.Staff([container])
        >>> abjad.show(container)

        ..  docs::

            {
                \time 3/4
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/extract_trivial_tuplets-moavfyqtxza.png

    ..  warning::

        The input selection must be a contiguous logical voice. When dealing
        with a container with multiple subcontainers (e.g. a score containing
        multiple staves), the best approach is to cycle through these
        subcontainers, applying this function to them individually.
    """
    if not isinstance(selection, abjad.Selection):
        raise TypeError("argument must be 'abjad.Selection'")

    tuplets = selection.tuplets()
    if len(tuplets) == 0:
        return

    for tuplet in tuplets:
        leaves = abjad.select(tuplet).leaves()
        if all(isinstance(leaf, abjad.Rest) for leaf in leaves) and len(leaves) > 1:
            duration = tuplet.multiplied_duration
            rests = abjad.LeafMaker()(None, duration)
            time_signature = abjad.get.indicator(
                leaves[0],
                abjad.TimeSignature,
            )
            if time_signature is not None:
                abjad.attach(time_signature, rests[0])
            abjad.mutate.replace(tuplet, rests)
        if abjad.get.sustained(tuplet):
            duration = tuplet.multiplied_duration
            n_elements = len(tuplet)
            after_tie = abjad.get.indicator(leaves[-1], abjad.Tie)
            for _ in range(n_elements - 1):
                tuplet.pop(-1)
            if not after_tie:
                abjad.detach(abjad.Tie, leaves[0])
            if duration.is_assignable:
                leaves[0].written_duration = duration
                abjad.mutate.extract(tuplet)
            elif duration.implied_prolation == 1:
                if isinstance(leaves[0], abjad.Note):
                    pitch = leaves[0].written_pitch
                elif isinstance(leaves[0], abjad.Chord):
                    pitch = leaves[0].written_pitches
                else:
                    pitch = None
                notes = abjad.LeafMaker()(pitch, duration)
                indicators = abjad.get.indicators(leaves[0])
                for indicator in indicators:
                    abjad.attach(indicator, notes[0])
                abjad.mutate.replace(leaves[0], notes)
                abjad.mutate.extract(tuplet)
            else:
                continue
    for tuplet in tuplets:
        if tuplet.trivializable():
            tuplet.trivialize()
            abjad.mutate.extract(tuplet)



"""
get
===

Auxjad's inspection functions. These can be accesses via either:

>>> abjad.get.auxjad_function(container)
>>> auxjad.get.auxjad_function(container)
"""

import abjad

from .leaves_are_tieable import leaves_are_tieable
from .rhythms_are_identical import rhythms_are_identical
from .selection_is_full import selection_is_full
from .selections_are_identical import selections_are_identical
from .time_signature_list import time_signature_list
from .underfull_duration import underfull_duration
from .virtual_fundamental import virtual_fundamental

### EXTENSION FUNCTIONS ###

abjad.get.leaves_are_tieable = leaves_are_tieable
abjad.get.rhythms_are_identical = rhythms_are_identical
abjad.get.selection_is_full = selection_is_full
abjad.get.selections_are_identical = selections_are_identical
abjad.get.time_signature_list = time_signature_list
abjad.get.underfull_duration = underfull_duration
abjad.get.virtual_fundamental = virtual_fundamental
